# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Evan Wojciechowski
# This software is part of projects sponsored by NASA and NSF (PI: V. Pankratius)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import matplotlib.pyplot as plt
import numpy as np
import pywt

from .vis_utils import *
    # block_output(), enable_output(), index_scale(),

# core functions

def calc_wp_deconstruct(calcData, wavelet = None):
    """
    simple function to calculate a wavelet deconstruction
    """

    if wavelet is None:
        wavelet = input('Wavelet not given, please enter a name (ex. haar): ').lower()

    #block_output() # necessary?
    deconPacket = pywt.WaveletPacket(data = calcData, wavelet = wavelet)
    #enable_output()

    return deconPacket

def calc_wp_reconstruct(deconPacket = None, calcData = None, wavelet = None, reconNodes = []):

    if deconPacket is None:
        if calcData is not None:
            deconPacket = calc_wp_deconstruct(calcData = calcData, wavelet = wavelet)
        else:
            print('Need one of deconPacket OR calcData')
            return

    reconPacket = pywt.WaveletPacket(data = None, wavelet = deconPacket.wavelet)

    if len(reconNodes) == 0:
        reconNodes = np.arange(1, deconPacket.maxlevel + 1)

    for n in reconNodes:
        reconPacket['a' * n] = deconPacket['a' * n]
        reconPacket['d' * n] = deconPacket['d' * n]

    reconPacket.reconstruct()

    return reconPacket


def plot_wp_deconstruct(deconPacket, deconNodes = [], mainTitle = 'Wavelet Deconstruction', plotRaw = True, show = True):

    col = 0

    if len(deconNodes) == 0:
        deconNodes = np.arange(1, deconPacket.maxlevel + 1)

    plt.figure(figsize = (12, 6.75))

    if plotRaw:
        plt.plot(deconPacket.data, color = coldict[col % 16], label = 'Raw')
        col += 1

    for n in deconNodes:
        indScaled = index_scale(toScale = np.arange(len(deconPacket['a' * n].data)), endRange = [0, len(deconPacket.data)])
        plt.plot(indScaled, deconPacket['a' * n].data / (np.sqrt(2) ** (n + 1)), color = coldict[col % 16], label = 'a.{}'.format(n))
        col += 1

    plt.legend(loc = 'best', fontsize = 12)
    plt.xticks(fontsize = 18)
    plt.yticks(fontsize = 18)
    plt.xlabel('Integer index from Raw (nodes scaled)', fontsize = 18)
    plt.ylabel('Amplitude', fontsize = 18)
    plt.title(mainTitle, fontsize = 18)

    if show:
        plt.show()

def plot_wp_showall(deconPacket, deconNodes = [], mainTitle = 'Individual Wavelet Nodes', plotRaw = True, show = True):

    col = 0

    if len(deconNodes) == 0:
        deconNodes = np.arange(1, deconPacket.maxlevel + 1)

    fig, axes = plt.subplots(len(deconNodes) + plotRaw, 1, sharex = True, figsize = (12, 6.75))

    if plotRaw:
        axes[0].plot(deconPacket.data, color = coldict[col % 16])
        axes[0].set_ylabel('Raw', fontsize = 18)
        col += 1

    for n in deconNodes:
        indScaled = index_scale(toScale = np.arange(len(deconPacket['a' * n].data)), endRange = [0, len(deconPacket.data)])
        axes[n - (not plotRaw)].plot(indScaled, deconPacket['a' * n].data / (np.sqrt(2) ** (n + 1)), color = coldict[col % 16])
        axes[n - (not plotRaw)].set_ylabel('a.{}'.format(n), fontsize = 18)
        col += 1

    for ax in axes:
        ax.tick_params(axis = 'x', labelsize = 18)
        ax.tick_params(axis = 'y', labelsize = 14)

    plt.xlabel('Integer index from Raw (nodes scaled)', fontsize = 18)
    axes[0].set_title(mainTitle, fontsize = 18)
    #plt.suptitle(mainTitle)

    if show:
        plt.show()
    

def plot_wp_reconstruct(reconPacket, calcData, mainTitle = 'Wavelet Reconstruction', plotRaw = True, show = True):

    col = 0

    plt.figure(figsize = (12, 6.75))

    plotArgs = calcData, reconPacket.data
    
    if plotRaw:
        plt.plot(calcData, color = coldict[col], label = 'Raw')
        col += 1

    plt.plot(reconPacket.data, color = coldict[col], label = 'Reconstruction ({} nodes)'.format(reconPacket.maxlevel))

    plt.title(mainTitle)
    plt.legend()

    if show:
        plt.show()


# wrappers

def run_plotWPDecon(inData, wavelet = None, deconNodes = [], mainTitle = 'Wavelet Deconstruction', plotRaw = True, show = True,):
    outData, outIndex = mod_data(inData)

    deconPacket = calc_wp_deconstruct(outData, wavelet = wavelet)

    plot_wp_deconstruct(deconPacket, deconNodes = deconNodes, mainTitle = mainTitle, plotRaw = plotRaw, show = show,)


def run_plotWPRecon(inData, wavelet = None, reconNodes = [], mainTitle = 'Individual Wavelet Nodes', plotRaw = True, show = True,):
    outData, outIndex = mod_data(inData)

    deconPacket = calc_wp_deconstruct(outData, wavelet = wavelet)
    reconPacket = calc_wp_reconstruct(deconPacket, calcData = inData, wavelet = wavelet, reconNodes = reconNodes)

    plot_wp_reconstruct(reconPacket, calcData = outData, mainTitle = mainTitle, plotRaw = plotRaw, show = show,)

def run_plotWPShowall(inData, wavelet = None, deconNodes = [], mainTitle = 'Wavelet Reconstruction', plotRaw = True, show = True):
    outData, outIndex = mod_data(inData)

    deconPacket = calc_wp_deconstruct(outData, wavelet = wavelet)

    plot_wp_showall(deconPacket, deconNodes = deconNodes, mainTitle = mainTitle, plotRaw = plotRaw, show = show)



#!/usr/bin/env python3

# The MIT License (MIT)
# Copyright (c) 2016 Massachusetts Institute of Technology
#
# Author: David Blair
# This software has been created in projects supported by the US National
# Science Foundation and NASA (PI: Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# """@package map_util
# A collection of map manipulation tools
# """

import numpy as np
import math
import copy
from scipy.optimize import minimize, brute


class Planet:
    """
    A class for storing variables about a planetary body
    """

    def __init__(self, name):
        '''
        Initialize Planet object

        @param name: The name of the planetary body choice of ('earth', 'wgs84', 'grs80', or 'moon'). 
                      'wgs84' and 'earth' provide the same planet.

        '''
        if (name.lower() == "earth") or (name.lower() == "wgs84"):
            self.a = 6378137.0
            self.b = 6356752.3142
            self.e_sq = (self.a**2 - self.b**2)/(self.a**2)
        elif name.lower() == "moon":
            self.a = 1738140.0
            self.b = 1735970.0
            #NB: The Moon is considered a sphere (e_sq = 0.0) by convention,
            #    although the true calculated value would be
            #    e_sq = 0.0024953633422124632
            self.e_sq = 0.0
        elif name.lower() == "grs80":
            self.a = 6378137.0
            self.b = 6356752.314140347
            self.e_sq = (self.a**2 - self.b**2)/(self.a**2)
	

        self.equator_1deg = np.pi*self.a/180
        self.avg_radius = np.mean((self.a,self.b))


    def get_lateraldist_array(self, ppd):
        """
        Retrieve the lateral distance array

        Get an array of the lateral size of 1/ppd of a degree of longitude at
        every 1/ppd of a degree of latitude. Results given in meters.

        Example: input of ppd = 1 for the body "Earth" results in an array 180
        cells long with lateraldist_array[90] = 111 (m).

        @param ppd: the number of pixels-per-degree-of-latitude; the resulting
                     array will therefore be (180*ppd) cells tall

        @returns lateraldist_array: an array of the size (in meters) of 1 degree of
                                    longitude at each 1/ppd-th of a degree of latitude
        """

        # Set up the parameters for the body (everything in meters)
        equator_1deg = np.pi * self.a/180

        # Set up the latitude array. Here we use np.linspace() instead of np.arange() because
        # we want to ensure that the resulting array has a length equal to the number of cells
        # of latitude in the dataset, and change the starting and ending values by (1/2)*(1/ppd)
        # so that each value is at the center-point of a cell in the dataset
        lat_array = np.linspace(-90 + (1/(2*ppd)),
                                90 - (1/(2*ppd)), num=(ppd*180), dtype=np.float)

        # Iterate over the latitudes & calculate the length of (1/ppd) degrees of longitude
        # at each one
        lateraldist_array = np.empty(len(lat_array))
        for i in range(len(lat_array)):

            spherical_approx = equator_1deg * np.cos(np.radians(lat_array[i]))
            ellipsoid_dist = spherical_approx / \
                             np.sqrt(1 - (self.e_sq * (np.sin(np.radians(lat_array[i])))**2))

            lateraldist_array[i] = ellipsoid_dist / ppd

            ###DEBUG
            #print("Spherical approximation at {0:3g}: {1:8g}".format(lat, spherical_approx))
            #print("  Ellipsoid calculation at {0:3g} degrees: {1:-12.2f}".format(lat, ellipsoid_dist))

        return lateraldist_array

    def get_lateraldist(self, lats, ppd):
        '''
        Get the lateral distance in meters for an input of lats

        @param lats: Either a scalar or an array of latitudes
        @param ppd: Pixels per degree of latitude
        
        @return Lateral distance at each latitude in meters
        '''
        # 1 degree distances east-west
        distances = wgs84_distance( (lats, np.zeros(len(lats))), (lats, np.ones(len(lats))/ppd), copy.copy(self) )

        # convert kilometers to meters
        distances *= 1000

        return distances

    def get_medialdist(self, lats, ppd):
        '''
        Get the medial distance at specific lattitudes

        @param lats: Either a scalar or an array of latitudes
        @param ppd: Pixels per degree of latitude
        
        @return Medial distance at each latitude in meters
        '''

        # Create list of lons at 0
        lons = np.zeros(len(lats))
        
        # 1 degree distances north-south
        distances = wgs84_distance( (lats, lons), (lats+(1/ppd), lons), copy.copy(self) )

        # convert kilometers to meters
        distances *= 1000

        return distances
        
def sanitize_latlon(lat_lon_tuple, ppd=1, start_from_90N=False):
    """
    Wraps around latitude & longitudes, including interpretation of points past the poles

    @param lat_lon_tuple: (lat, lon), in either degrees or pixels
    @param ppd: pixels-per-degree
    @param start_from_90N: consider 90N to be 0 latitude

    @return Latitude and Longitude after they have been sanitized
    """

    lat, lon = lat_lon_tuple

    if start_from_90N:
        if lat < 0:
            return [-lat, (lon + 180*ppd)%(360*ppd)]
        elif lat > 180*ppd:
            return [-360*ppd - lat, (lon + 180*ppd)%(360*ppd)]
        else:
            return [lat, lon%(360*ppd)]
    else:
        if lat > 90*ppd:
            return [180*ppd - lat, (lon + 180*ppd)%(360*ppd)]
        elif lat < -90*ppd:
            return [-180*ppd - lat, (lon + 180*ppd)%(360*ppd)]
        else:
            return [lat, lon%(360*ppd)]


def trim_map(array, ppd, nswe, lat_npole=90, lon_offset=0):
    """
    Returns a copy of a map/array trimmed to the given N, S, W, E extents

    @param array: the input array to be trimmed
    @param ppd: the pixels-per-degree of the array
    @param nswe: a 1x4 array of the desired [N, S, W, E] edges
    @param lat_npole: the latitude of the N Pole in the same system as the
                          given N, S, W, E values
    @param lon_offset: the longitude of the prime meridian in the same system
                          as the given N, S, W, E values

    @returns trimmed_map: the input data trimmed to the desired edges
    """

    map_N = int((lat_npole - nswe[0])*ppd)
    map_S = int((lat_npole - nswe[1])*ppd)
    map_W = int((nswe[2] + lon_offset)*ppd)
    map_E = int((nswe[3] + lon_offset)*ppd)
    trimmed_map = array[map_N:map_S, map_W:map_E]

    return trimmed_map


def calc_slopes(topo_array, ppd, planet, scaled=True,
                nswe="global", lon_offset=0, lat_npole=90):
    """
    Calculate a slope map from a topographic dataset

    For now, this tool assumes a global topographic dataset; in the future, it
    will be expanded to work on regional datasets as well

    @param topo_array: a global topographic dataset, in numpy array form
    @param ppd: the pixels-per-degree of the topo array
    @param planet: The planetary body in question
    @param scaled: whether values should be scaled by latitude
    @param nswe: the (NW,SE) corners of the area-of-interest
    @param lon_offset: the longitude of the prime meridian in the same system
                          as the given N, S, W, E values
    @param lat_npole: the latitude of the N Pole in the same system as the
                          given N, S, W, E values
    """

    # Parameters for different celestial bodies
    dist_1px_lat = (2 * np.pi * planet.avg_radius)/(360*ppd)

    ###DEBUG: OK
    #print("1/ppd of 1 degree around the lunar circumference:", 10921000 / 360 / ppd)
    #print("calculated as:", dist_1px_lat)

    # Calculate only part of the array?
    if nswe != "global":

        # Convert lat/lon values to pixel (cell) coordinates for slicing
        #N = (lat_npole - nswe[0])*ppd
        #S = (lat_npole - nswe[1])*ppd
        #W = (nswe[2] + lon_offset)*ppd
        #E = (nswe[3] + lon_offset)*ppd

        ###DEBUG
        #print(W,E,S,N)

        # Taking slice of data from [N:S, W:E]
        topo_array = trim_map(topo_array, ppd, nswe, lat_npole, lon_offset)

    # Get the array of the size of 1 pixel East-West at the ppd of the data. If
    # the data is non-global, trim the array to fit.
    lateraldist_array = planet.get_lateraldist_array(ppd)
    if nswe != "global":
        lateraldist_array = lateraldist_array[(lat_npole - nswe[0])*ppd:
                                              (lat_npole - nswe[1])*ppd]
        #lateraldist_array = lateraldist_array[N:S]

    ###DEBUG
    #print("max of lateraldist_array:", max(lateraldist_array))
    #print(lateraldist_array.shape)

    # Use np.gradient() to get the by-rows (North-South) and by-columns (East-West)
    # components of the slope very quickly
    S_slope, E_slope = np.gradient(topo_array)

    # Since numpy has no idea of the physical scale, though, we have to multiply
    # by something so that we have a distance/distance measurement. If we're
    # scaling with latitude, we can do this intelligently, otherwise we'll just
    # multiply by the distance along a great circle
    if scaled:
        E_slope = E_slope / lateraldist_array[:, np.newaxis]
        S_slope = S_slope / dist_1px_lat
    else:
        E_slope = E_slope / dist_1px_lat
        S_slope = S_slope / dist_1px_lat

    # Since each cell in the topo & slope arrays represents 1 pixel, and each
    # pixel represents a plane, the overall slope of that plane can be calculated
    # by taking the root of the sum of the squares of the slope components
    slope_array = np.sqrt(E_slope**2 + S_slope**2)
    # TODO: is this correct?

    # Calculate the aspect array (direction of the slope)
    aspect_array = np.degrees(np.arctan2(-E_slope, S_slope))

    # Go from dist/dist to degrees
    slope_array = np.degrees(np.arctan(slope_array))

    ###DEBUG:
    #f, [[ax1,ax4],[ax2,ax3]] = plt.subplots(2,2)
    #f1 = ax1.imshow(topo_array)
    #div1 = make_axes_locatable(ax1)
    #cax1 = div1.append_axes("right", size="15%", pad=0.1)
    #plt.colorbar(f1, cax=cax1)
    #f2 = ax2.imshow(E_slope, cmap="RdBu")
    #div2 = make_axes_locatable(ax2)
    #cax2 = div2.append_axes("right", size="15%", pad=0.1)
    #plt.colorbar(f2, cax=cax2)
    #f3 = ax3.imshow(S_slope, cmap="RdBu")
    #div3 = make_axes_locatable(ax3)
    #cax3 = div3.append_axes("right", size="15%", pad=0.1)
    #plt.colorbar(f3, cax=cax3)
    #f4 = ax4.imshow(aspect_array, cmap="gray")
    #div4 = make_axes_locatable(ax4)
    #cax4 = div4.append_axes("right", size="15%", pad=0.1)
    #plt.colorbar(f4, cax=cax4)

    return slope_array



def wgs84_distance(point1, point2, planet=Planet("wgs84"), miles=False):
    """
    Vincenty distance adapted from public domain vincenty package

    Adapted from https://github.com/maurycyp/vincenty
    @verbatim
    Vincenty's formula (inverse method) to calculate the distance (in
    kilometers or miles) between two points on the surface of a spheroid
    >>> wgs84_distance((0.0, 0.0), (0.0, 0.0))  # coincident points
    0.0
    >>> wgs84_distance((0.0, 0.0), (0.0, 1.0))
    111.319491
    >>> wgs84_distance((0.0, 0.0), (1.0, 0.0))
    110.574389
    >>> wgs84_distance((0.0, 0.0), (0.5, 179.5))  # slow convergence
    19936.288579
    >>> wgs84_distance((0.0, 0.0), (0.5, 179.7))  # failure to converge
    >>> boston = (42.3541165, -71.0693514)
    >>> newyork = (40.7791472, -73.9680804)
    >>> wgs84_distance(boston, newyork)
    298.396057
    >>> wgs84_distance(boston, newyork, miles=True)
    185.414657
    @endverbatim

    @param point1: (lat1, lon1)
    @param point2: (lat2, lon2)
    @param planet: Planet to perform the computation on
    @param miles: Convert result to miles (default kilometers)

    @return distance between point1 and point2
    """
    
    MILES_PER_KILOMETER = 0.621371
    MAX_ITERATIONS = 200
    CONVERGENCE_THRESHOLD = 1e-12  # .000,000,000,001

    a = planet.a
    b = planet.b
    f = 1 - b / a
        
    def calc_distance(lat1, lon1,lat2, lon2):
        # short-circuit coincident points
        if lat1 == lat2 and lon1 == lon2:
            return 0.0

        U1 = math.atan((1 - f) * math.tan(math.radians(lat1)))
        U2 = math.atan((1 - f) * math.tan(math.radians(lat2)))
        L = math.radians(lon2 - lon1)
        Lambda = L

        sinU1 = math.sin(U1)
        cosU1 = math.cos(U1)
        sinU2 = math.sin(U2)
        cosU2 = math.cos(U2)

        for iteration in range(MAX_ITERATIONS):
            sinLambda = math.sin(Lambda)
            cosLambda = math.cos(Lambda)
            sinSigma = math.sqrt((cosU2 * sinLambda) ** 2 +
                                 (cosU1 * sinU2 - sinU1 * cosU2 * cosLambda) ** 2)
            if sinSigma == 0:
                return 0.0  # coincident points
            cosSigma = sinU1 * sinU2 + cosU1 * cosU2 * cosLambda
            sigma = math.atan2(sinSigma, cosSigma)
            sinAlpha = cosU1 * cosU2 * sinLambda / sinSigma
            cosSqAlpha = 1 - sinAlpha ** 2
            try:
                cos2SigmaM = cosSigma - 2 * sinU1 * sinU2 / cosSqAlpha
            except ZeroDivisionError:
                cos2SigmaM = 0
            C = f / 16 * cosSqAlpha * (4 + f * (4 - 3 * cosSqAlpha))
            LambdaPrev = Lambda
            Lambda = L + (1 - C) * f * sinAlpha * (sigma + C * sinSigma *
                                                   (cos2SigmaM + C * cosSigma *
                                                    (-1 + 2 * cos2SigmaM ** 2)))
            if abs(Lambda - LambdaPrev) < CONVERGENCE_THRESHOLD:
                break  # successful convergence
        else:
            return None  # failure to converge

        uSq = cosSqAlpha * (a ** 2 - b ** 2) / (b ** 2)
        A = 1 + uSq / 16384 * (4096 + uSq * (-768 + uSq * (320 - 175 * uSq)))
        B = uSq / 1024 * (256 + uSq * (-128 + uSq * (74 - 47 * uSq)))
        deltaSigma = B * sinSigma * (cos2SigmaM + B / 4 * (cosSigma *
                     (-1 + 2 * cos2SigmaM ** 2) - B / 6 * cos2SigmaM *
                     (-3 + 4 * sinSigma ** 2) * (-3 + 4 * cos2SigmaM ** 2)))
        s = b * A * (sigma - deltaSigma)

        s /= 1000  # meters to kilometers
        if miles:
            s *= MILES_PER_KILOMETER  # kilometers to miles

        return round(s, 6)

    lat1 = point1[0]
    lon1 = point1[1]

    lat2 = point2[0]
    lon2 = point2[1]

    if np.isscalar(lat1) and np.isscalar(lon1) and np.isscalar(lat2) and np.isscalar(lon2) :
        return calc_distance(lat1, lon1, lat2, lon2)
    else:
        return np.fromiter(map(lambda x: calc_distance(*x), zip(lat1, lon1, lat2, lon2)),
                           dtype=np.float,count=np.max([len(lat1), len(lon1), len(lat2), len(lon2)]))

    
        
# This function converts pixel x,y coordinates into global coordinates
def global_coords(x_in,y_in,coeffs):
    '''
    Transform pixel coordinates into global coords using affine transformation coefficients

    @param x_in: X pixel coordinates
    @param y_in: Y pixel coordinates
    @param coeffs: Affine transformation coefficients

    @return global coordinates
    '''
    
    x = x_in + 0.5
    y = y_in + 0.5
    return (coeffs[0] + coeffs[1]*x + coeffs[2]*y, coeffs[3] + coeffs[4]*x + coeffs[5]*y)


def gps_to_pixel(gpsmethod, gps_coord, bounds):
    """
    Function for finding the pixel coordinate associated with a gps coordinate

    @param gpsmethod: GPS coordinate mapping function from above
    @param gps_coord: GPS coordinate to match, as (lat,lon)
    @param bounds: Pixel bounds to search within ((y_low,y_high),(x_low,x_high))

    @return Nearest integer pixel value

    """

    # map_ops great circle distance
    func = lambda imgxy: wgs84_distance(gpsmethod(imgxy[0], imgxy[1]), (gps_coord[0], gps_coord[1]))
    res = brute(func, bounds)

    # return np.array([int(round(cc)) for cc in res])
    return np.array(res)

#### Test Examples #############################################################

# Testing sanitize_latlon
#print(sanitize_latlon((0,0)))
#print(sanitize_latlon((30,362)))
#print(sanitize_latlon((30,-40)))
#print(sanitize_latlon((92,0)))
#print()
#print(sanitize_latlon((-2,0), start_from_90N=True))
#print(sanitize_latlon((-2,-2), start_from_90N=True))
#print()
#ppd = 16
#print(sanitize_latlon((-2*ppd,0), ppd=ppd, start_from_90N=True))
#print(sanitize_latlon((-2*ppd, -2*ppd), ppd=ppd, start_from_90N=True))


#nw = [(90 - -42)*ppd, (345)*ppd]
#se = [(90 - -45)*ppd, (352)*ppd]
#topo_subset = data_array[nw[0]:se[0],nw[1]:se[1]]
#
#local_slope_map = calc_slopes(topo_subset, ppd)
#plt.imshow(local_slope_map, cmap="jet")
#plt.savefig("tycho.jpg")
#plt.show()


#ppd = int(topo_array.shape[0] / 180)
# Tycho
#nw = [131, 348]
#se = [135, 352]
#nwse = [nw,se]
# Tiny
#nw = [88, 178]
#se = [92, 182]
#nwse = [nw,se]
# Very large area
#nw = [10,90]
#se = [170,270]
#nwse = [nw,se]

#map1 = calc_slopes(topo_array, ppd, nwse=nwse, scaled=False)
#map2 = calc_slopes(topo_array, ppd, nwse=nwse, scaled=True)



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


class PipelineItem:
    '''
    The general class used to create pipeline items.
    '''

    def __init__(self, str_description, ap_paramList=[]):
        '''
        Initialize an object

        @param str_description: String description of filter
        @param ap_paramList: List of AutoParam parameters.
        '''
        self.str_description = str_description
        self.ap_paramList = ap_paramList
        self.ap_paramNames = []

    def perturbParams(self):
        '''choose other random value for all parameters'''
        for param in self.ap_paramList:
            param.perturb()
            
    def resetParams(self):
        '''set all parameters to initial value'''
        for param in self.ap_paramList:
            param.reset()
            
    def process(self, obj_data):
        '''
        The actual filter processing. Empty in this generic filter. 

        @param obj_data: Data wrapper that will be processed
        '''
        pass
 
    def __str__(self):
        ''' 
        String represntation of object.

        @return String listing all currenter parameters
        '''
        return str([str(p) for p in self.ap_paramList])
        
    def getMetadata(self):
        ''' 
        Retrieve metadata about filter
        
        @return String containing the item description and current parameters for filter.
        '''
        return self.str_description + str([str(p) for p in self.ap_paramList])




# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.



# Framework import
from skdiscovery.data_structure.framework.base import PipelineItem

# 3rd party libraries import
import pandas as pd


class CombineColumns(PipelineItem):
    '''
    Create a new column by selecting data from a column

    Fills in any missing values using a second column
    '''
    def __init__(self, str_description, column_1, column_2, new_column_name):
        '''
        Initialize a CombineColumns object

        @param str_description: String describing filter
        @param column_1: Name of primary column
        @param column_2: Name of secondary column to be used
        when data from the primary column is not avaiable
        @param new_column_name: Name of resulting column
        '''
        self.column_1 = column_1
        self.column_2 = column_2
        self.new_column_name = new_column_name
        
        super(CombineColumns,self).__init__(str_description)
    def process(self, obj_data):
        ''' 
        Apply combine column filter to data set, operating on the data_obj
        
        @param obj_data: Table data wrapper.
        '''
        for label, data in obj_data.getIterator():
            if self.column_1 in data.columns and self.column_2 in data.columns:
                 # replacing all null median data with mean data
                col1_null_index = pd.isnull(data.loc[:,self.column_1])


                data.loc[:,self.new_column_name] = data.loc[:,self.column_1]

                # Check if there is any replacement data available
                if (~pd.isnull(data.loc[col1_null_index, self.column_2])).sum() > 0:
                    data.loc[col1_null_index, self.new_column_name] = data.loc[col1_null_index, self.column_2]
                
                
            elif self.column_2 in data.columns and self.column_1 not in data.columns:
                data.loc[:,self.new_column_name] = data.loc[:,self.column_2]

            elif self.column_2 not in data.columns and self.column_1 in data.columns:
                data.loc[:,self.new_column_name] = data.loc[:,self.column_1]

            else:
                raise KeyError('data needs either "' + self.column_2 + '" or "' + self.column_1 + '" or both')



# The following license applies to angular_seperation, abs_mag,
# app_mag, and move_point

# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# The functions z_to_v, v_to_z, lf, dlf, cdf_dlf, and inv_cdf_dlf are
# adapted from R code written by Cody Rude. The function NFW was
# adapted from python code written by Cody Rude. The original versions
# of these are licensed under the MIT license below. Any changes made
# to the original version are copyright MIT and follow the license and
# copyright listed above.


# The MIT License (MIT)
# Copyright (c) 2016 Cody Rude
#
# Author: Cody Rude
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import numpy as np
import math
from astropy.constants import c
from astropy import units as u
from astropy.coordinates import SkyCoord as SC
from astropy.cosmology import FlatLambdaCDM

from scipy.optimize import root
from scipy.integrate import quad


def z_to_v(z):
    '''
    Convert redshift to km/s assuming shift is due to velocity using special relativity

    @param z: Redshift

    @return speed in km/s assuming shift is due to motion using special relativity
    '''
    ckms = c.to('km/s').value
    return (ckms * ( (z+1)**2 - 1) / ( (z+1)**2 + 1 ))


def v_to_z(v):
    '''
    Convert km/s to redshift assuming all are using special relativity

    @param v: velocity in km/s

    @return Redshift of object with speed in km/s
    '''
    ckms = c.to('km/s').value    
    return np.sqrt((1 + v/ckms) / (1 - v/ckms)) - 1


def angular_separation(ra1, dec1, ra2, dec2):
    '''
    Angular seperation between two objects via the haversine formula. 
    
    All inputs are in degrees.

    Formula obtained from http://www.movable-type.co.uk/scripts/gis-faq-5.1.html

    Formula originally presented in
    R.W. Sinnott, "Virtues of the Haversine", Sky and Telescope, vol. 68, no. 2, 1984, p. 159

    @param ra1: Right Ascention of first object (degrees)
    @param dec1: Declination of first object (degrees)
    @param ra2: Right Ascention of second object (degrees)
    @param dec2: Declination of second object (degrees)

    @return angular seperation between two objects
    '''

    ra1 = ra1 * np.pi / 180.
    dec1 = dec1 * np.pi / 180.
    ra2 = ra2 * np.pi / 180.
    dec2 = dec2 * np.pi / 180.

    dra  = ra2 - ra1
    ddec = dec2 - dec1

    a = np.sin(ddec/2)**2 + np.cos(dec1) * np.cos(dec2) * np.sin(dra/2)**2
    return 2*np.arcsin(np.minimum(1,np.sqrt(a))) * 180. / np.pi
    


def move_point(ra,dec,ang_dist,bearing):
    '''
    Move a point along a great circle at a particular bearing

    All inputs are in degrees
    The formula was obtained from http://www.movable-type.co.uk/scripts/latlong.html

    @param ra: Starting right ascension
    @param dec: Starting declination
    @param ang_dist: Angular distance to travel
    @param bearing: Direction to travel (0 is north, 90 is positive RA)

    @return tuple containing updated ra and dec
    '''
    
    ra = ra * np.pi / 180.
    dec = dec * np.pi / 180.
    ang_dist = ang_dist * np.pi / 180.
    bearing = bearing * np.pi / 180
    
    new_dec = np.arcsin( np.sin(dec) * np.cos(ang_dist) + np.cos(dec)*np.sin(ang_dist)*np.cos(bearing))
    new_ra = ra + np.arctan2(np.sin(bearing) * np.sin(ang_dist) * np.cos(dec), 
                             np.cos(ang_dist) - np.sin(dec)*np.sin(new_dec))
    
    new_dec *= 180. / np.pi
    new_ra *= 180. / np.pi
    
    new_ra %= 360.
    
    return new_ra, new_dec
    

def abs_mag(app_mag, z):
    ''' 
    Get the absolute magnitude from apparent magnitude
    
    Assumes concordance cosmology. No kcorrection is applied.

    @param app_mag: Apparent magnitude
    @param z: Redshift

    @return absolute magnitude of object at z
    '''
    
    cosmo = FlatLambdaCDM(H0=70, Om0=0.3)
    distmod = 5*np.log10(cosmo.luminosity_distance(z).to('pc').value)-5
    return app_mag - distmod


def app_mag(abs_mag, z):
    ''' 
    Get the apparent magnitude from absolute magnitude.

    Assumes concordance cosmology. No kcorrection is assumed.

    @param abs_mag: Absolute magnitude
    @param z: Redshift

    @return apparent magnitude of object at z
    '''

    cosmo = FlatLambdaCDM(H0=70, Om0=0.3)
    distmod = 5*np.log10(cosmo.luminosity_distance(z).to('pc').value)-5
    return abs_mag + distmod


def nfw(R, norm_constant, Rs, Rcore):
    '''
    2D Navarro-Frenk-White surface radial profile probability density

    See:
    Navarro, J. F., Frenk, C. S., & White, S. D. M. 1996, ApJ, 462, 563
    Bartelmann, M., A&A, 1996, 313, 697
    Rykoff, E.S., et al., ApJ, 746, 178

    @param R: Radius
    @param norm_constant: Normalization constant
    @param Rs: Scale radius
    @param Rcore: Since NFW profile diverges at R=0, the value at the center is held fixed starting at Rcore

    @return probability density of profile at R
    '''

    def compute_nfw(R):
        if R < Rcore:
            R2 = Rcore
        else:
            R2 = R

        if (R2/Rs) < 0.999:
            func = (1 - 2 * math.atanh(math.sqrt((1 - R2/Rs) / (R2/Rs + 1))) / math.sqrt(1 - (R2/Rs)**2)) / ((R2/Rs)**2 - 1)

        # There are some computational issues as R2/Rs -> 1, using taylor expansion of the function
        # around this point
        elif (R2/Rs) < 1.001:
            func = -(20/63)*((R2/Rs)**3 - 1) + (13/35)*((R2/Rs)**2 - 1) - (2/5)*(R2/Rs) + (11/15)

        else:
            func = (1 - 2 * math.atan(math.sqrt((R2/Rs - 1) / (R2/Rs + 1))) / math.sqrt((R2/Rs)**2 - 1)) / ((R2/Rs)**2 - 1)


        return norm_constant * 2 * math.pi * R * func

    if np.isscalar(R):
        return compute_nfw(R)

    else:
        return np.fromiter(map(compute_nfw, R), np.float, count = len(R))


def lf(x, A, mstar, alpha):
    ''' 
    Schechter function

    @param x: magnitude
    @param A: Scale factor
    @param mstar: Knee of distribution
    @param alpha: Faint-end turnover
    
    @return float: Schecter function at magnitude x
    '''
    return(A * 10**(-0.4 * (x - mstar) * (alpha + 1)) * np.exp(-10**(-0.4 * (x - mstar)) ))


def dlf(x, A, m1, a1, m2, a2):
    ''' 
    double Schechter function. Second LF is set to be 2*A of first LF.

    @param x: magnitude
    @param A: Scale factor
    @param m1: Knee of distribution 1
    @param a1: Faint-end turnover of first lf
    @param m2: Knee of distribution 2
    @param a2: Faint-end turnover of second lf
    
    @return float: Double Schecter function at magnitude x
    '''
    
    return(lf(x, A, m1, a1) + lf(x, 2*A, m2, a2))


def cdf_dlf(x, A, m1, a1, m2, a2, start=-26):
    ''' 
    Cumulative  Schechter function. Second LF is set to be 2*A of first LF.

    @param x: magnitude
    @param A: Scale factor
    @param m1: Knee of distribution 1
    @param a1: Faint-end turnover of first lf
    @param m2: Knee of distribution 2
    @param a2: Faint-end turnover of second lf
    @param start: Brightest magnitude
    
    @return Probability that galaxy has a magnitude greater than x
    '''
    def integrate(in_x):
        return quad(dlf, start,in_x,args=(A,m1,a1,m2,a2))[0]

    if np.isscalar(x):
        x = np.array([x])

    return np.fromiter(map(integrate,x),np.float,count=len(x))


def inv_cdf_dlf(p, A, m1, a1, m2, a2, start=-26, end=-15):

    ''' 
    Inverse Cumulative Schechter function. Second LF is set to be 2*A of first LF.

    @param p: probability
    @param A: Scale factor
    @param m1: Knee of distribution 1
    @param a1: Faint-end turnover of first lf
    @param m2: Knee of distribution 2
    @param a2: Faint-end turnover of second lf
    @param start: Brightest magnitude
    @param end: Faintest possible magnitude
    
    @return Magnitude associated with cdf probability p
    '''
    def get_root(p):
        return root(lambda x: cdf_dlf(x,A,m1,a1,m2,a2,start)-p, (start + end)/2).x[0]

    if np.isscalar(p):
        return get_root(p)
    else:
        return np.fromiter(map(get_root,p),np.float,count=len(p))



# The MIT License (MIT)
# Copyright (c) 2015 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from skdiscovery.data_structure.framework import PipelineItem
import numpy as np
from sklearn.decomposition import PCA
from sklearn.decomposition import FastICA


class General_Component_Analysis(PipelineItem):
    ''' 
    Performs a general component analysis on table data.

    Currently, the two built-in types of analysis are either ICA or PCA.
    '''

    def __init__(self, str_description, ap_paramList, n_components, column_names):
        '''
        Initialize Analysis object

        @param str_description: String description of analysis
        @param ap_paramList[component_type]: Type of CA; either PCA or ICA
        @param ap_paramList[start_time]: Starting time for CA
        @param ap_paramList[end_time]: ending time for CA
        @param n_components: Number of components to compute
        @param column_names: Columns names to use 
        '''
        
        self.str_description = str_description
        self.ap_paramList = ap_paramList
        self.ap_paramNames = ['component_type','start_time','end_time']
        self.n_components = n_components
        self.column_names = column_names
        self.results = dict()

    
    def process(self, obj_data):
        ''' 
        Perform component analysis on data

        Results are added to the data wrapper as a dictionary with
        results['CA'] = Eigenvenctors
        results['Projection'] = Projection on to the eigenvectors
        
        @param obj_data: Data wrapper
        '''

        component_type = self.ap_paramList[0]()
        start_time = self.ap_paramList[1]()
        end_time = self.ap_paramList[2]()

        num_components = self.n_components

        results = dict()
        results['start_date'] = start_time
        results['end_date'] = end_time

        cut_data = []
        label_list = []
        for label, data  in obj_data.getIterator():
            for column in self.column_names:
                cut_data.append(data.loc[start_time:end_time, column])
                label_list.append(label)

        cut_data = np.array(cut_data)

        if len(cut_data) > 0:
            if component_type == 'ICA' :
                ca = FastICA(n_components = num_components)
            else:
                ca = PCA(n_components = num_components)
            
                time_projection = ca.fit_transform(cut_data.T)
                results['CA'] = ca
                results['Projection'] = time_projection

        else:
            results['CA'] = None
            results['Projection'] = None

        results['labels'] = label_list

        obj_data.addResult(self.str_description, results)



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy as sp

from skdaccess.utilities import pbo_util
from skdiscovery.data_structure.series.analysis.mogi import MogiVectors

from skdiscovery.utilities.patterns import pbo_tools


def calc_distance_map(pipeline, ap_name, ca_name, ca_type ,plotFlag = True, histIdx = False, fontsize=10):
    """ 
    Calculates distances/similarities between pipeline runs

    Optionally visualizes the result as a seaborn clustermap for PBO
    pipelines (requires multiple stations)

    Calculates the square root of the summed squared differences between eigenvectors.
    Only works, because of internal assumptions, on pipelines with multiple stations
    Returns the distances as a pandas dataframe
    
    @param pipeline: Pipeline to analyze.
    @param ap_name: Name of the pipeline item that is being perturbed
    @param ca_name: Name of the pipeline item used as the comparison metric for calculating the distance
    @param ca_type: Type of comparison metric [PCA for PCA, MogiSource of Mogi Source, MogiVector for Mogi vectors]
    @param plotFlag: Boolean flag for plotting the clustermap of distances
    @param histIdx: Flag for returning the perturbed pipeline item parameters
    @param fontsize: Fontsize adjustments

    @return cg: The generated clustermap of the calculated distances/similarities
    @return dist_mat: A matrix of the calculated distances/similarities
    @return history: The record of the perturbed pipeline item parameters
    """
    # a history of the perturbed pipeline item
    history = []
    for runInfo in pipeline.getMetadataHistory():
        for stageItem in runInfo:
            if ap_name in stageItem:
                history.append(stageItem.rsplit(ap_name)[1].strip('[]'))
        
    # number of runs
    num_results = len(pipeline.RA_results)  
    dist_mat = np.zeros([num_results, num_results])
        
    # compute distances between all pairs of runs
    for i in range(num_results):
        for j in range(i, num_results):       
            # Check the ca_name type to properly format the type of comparison
            if ca_type == 'PCA':
                ctitle='PCA Vector Distance Similarity'
                summation = 0
                rstation_list = pipeline.RA_results[i][ca_name]['labels']
                num_stations = len(rstation_list)
                # some redundancy because of the way the modules/functions were structured
                # eigenvectors (lat and lon) for the one run
                coord_list = pbo_util.getStationCoords(pipeline.data_fetcher.meta_data, rstation_list)
                _,_,EV1_lat,EV1_lon,_ = pbo_tools.dirEigenvectors(coord_list, pipeline.RA_results[i][ca_name]['CA'].components_[0])
                # and for the other run
                _,_,EV2_lat,EV2_lon,_ = pbo_tools.dirEigenvectors(coord_list, pipeline.RA_results[j][ca_name]['CA'].components_[0])
                
                for k in range(num_stations):
                    ev1 = np.hstack((EV1_lat[k], EV1_lon[k]))
                    ev2 = np.hstack((EV2_lat[k], EV2_lon[k]))
                    # calculate the euclidean distance difference at each station
                    summation += sp.spatial.distance.euclidean(ev1, ev2)**2
                
            elif ca_type == 'MogiSource':
                ctitle='Mogi Source Distance [deg+km]'
                # for now, just uses lat, lon, and depth for Mogi comparison
                ev1 = np.array([pipeline.RA_results[i][ca_name]['lat'], pipeline.RA_results[i][ca_name]['lon'], pipeline.RA_results[i][ca_name]['depth']])
                ev2 = np.array([pipeline.RA_results[j][ca_name]['lat'], pipeline.RA_results[j][ca_name]['lon'], pipeline.RA_results[j][ca_name]['depth']])
                summation = sp.spatial.distance.euclidean(ev1, ev2)**2
            elif ca_type == 'MogiVector':
                ctitle='Summed Mogi Vector Distance [mm]'
                # do the same comparison as eigenvectors for the mogi modeled vectors
                summation = 0
                rstation_list = pipeline.RA_results[i][ca_name]['labels']
                num_stations = len(rstation_list)
                # got mogi vectors for the two runs
                coord_list = np.array(pbo_util.getStationCoords(pipeline.data_fetcher.meta_data, rstation_list))
                mogi_x_1, mogi_y_1 = MogiVectors(pipeline.RA_results[i][ca_name],coord_list[:,0],coord_list[:,1])
                mogi_x_2, mogi_y_2 = MogiVectors(pipeline.RA_results[j][ca_name],coord_list[:,0],coord_list[:,1])
                
                for k in range(num_stations):
                    ev1 = np.hstack((mogi_x_1[k], mogi_y_1[k]))
                    ev2 = np.hstack((mogi_x_2[k], mogi_y_2[k]))
                    # calculate the euclidean distance difference at each station
                    summation += sp.spatial.distance.euclidean(ev1, ev2)**2
                # as all pca amplitudes are the same, to scale Mogi to difference in mm
                summation *= (pipeline.RA_results[0][ca_name]['pca_amplitude'])**2
                
            dist_mat[i][j] = np.sqrt(summation)

    dist_mat += dist_mat.transpose()
    if histIdx:
        dist_mat = pd.DataFrame(dist_mat, index=['Configuration '+str(ii).zfill(2) for ii in range(len(history))], columns=['Configuration '+str(ii).zfill(2) for ii in range(len(history))])
    else:
        dist_mat = pd.DataFrame(dist_mat, index=history, columns=history)
    
    if plotFlag:
        # Importing seaborn here in order to
        # return matplotlib to it's original style
        import seaborn as sns
        sns.set(font='serif')

        cg = sns.clustermap(dist_mat)
        plt.setp(cg.ax_heatmap.yaxis.get_majorticklabels(), rotation=0, fontsize=fontsize);
        plt.setp(cg.ax_heatmap.xaxis.get_majorticklabels(), rotation=90, fontsize=fontsize);
        cg.cax.set_title(ctitle,fontsize=fontsize);
        sns.reset_orig()

        if histIdx:
            return cg, history
        else:
            return cg
    else:
        if histIdx:
            return dist_mat, history
        else:
            return dist_mat



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and
# NASA AIST Project "Computer-Aided Discovery of Earth Surface
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from skdiscovery.data_structure.framework import PipelineItem
from skdaccess.utilities.kepler_util import normalize

class NormalizeFilter(PipelineItem):
    '''
    Normalize data using median filter
    '''


    def __init__(self, str_description, column='PDCSAP_FLUX', group_column = 'QUARTER'):
        '''
        Initilaize NormalizeFilter

        @param str_description: String describing filter
        @param column: Name of column to normalize
        @param group_column: Column to use to group data
        '''

        self.column = column
        self.group_column = group_column

        super(NormalizeFilter, self).__init__(str_description)



    def process(self, obj_data):
        '''
        Apply Normalization filter to data wrapper

        @param obj_data: Input table data wrapper
        '''

        for label, data in obj_data.getIterator():
            normalize(data, column = self.column, group_column = self.group_column)






# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and
# NASA AIST Project "Computer-Aided Discovery of Earth Surface
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# """@package pbo_tools
# Tools for working with PBO GPS data
# """

import time

import numpy as np
import pandas as pd

import skdiscovery.utilities.planetary.map_util as mo

def mogi(xdata, y, x, source_depth, amplitude, latlon=True):
    '''
    Compute the surface deformation due to changes in a mogi source

    @param xdata: List of the position data with each array element containing [ direction (x, y, or z), lat, lon ]
    @param y: Source y Position of (default: latitude)
    @param x: Source x Position (default longitude)
    @param source_depth: Depth of source
    @param amplitude: Amplitude of mogi source
    @param latlon: Source y is latitude and source x is longitude

    @return list of resulting deformation for each point in xdata
    '''
    source_coords = (y, x)

    results = []

    for data in xdata:

        dim = data[0]
        station_coords = (float(data[1]),float(data[2]))
        # print(station_coords)

        if latlon==True:
            y_distance = mo.wgs84_distance( source_coords, (station_coords[0], source_coords[1]) )
            x_distance = mo.wgs84_distance( source_coords, (source_coords[0], station_coords[1]) )
            x_distance = x_distance * np.sign(station_coords[1] - source_coords[1])
            y_distance = y_distance * np.sign(station_coords[0] - source_coords[0])
        else:
            y_distance = station_coords[0] - source_coords[0]
            x_distance = station_coords[1] - source_coords[1]

        R3 = (x_distance**2 + y_distance**2 + source_depth**2)**(3/2)

        result = None

        if dim == 'x':
            result = amplitude * x_distance / R3
        elif dim == 'y':
            result = amplitude * y_distance / R3
        elif dim == 'z':
            result = amplitude * source_depth / R3
        else:
            print("Did not understand dimension")

        results.append(result)
    return results



def finite_sphere(xdata, lat, lon, source_depth, amplitude, alpha_rad):
    '''
    Compute the surface deformation due to changes in a finite sphere source

    For reference, see "Volcano Deformation", Dzurisin 2006, pg 290
    (http://link.springer.com/book/10.1007/978-3-540-49302-0)

    @param xdata: List of the position data with each array element containing [ direction (x, y, or z), lat, lon ]
    @param lat: Latitude of source
    @param lon: Longitude of source
    @param source_depth: Depth of source
    @param amplitude: Ampltiude of source
    @param alpha_rad: Alpha radius of the source

    @return list of resulting deformation for each point in xdata

    '''
    nu_v = .25
    C1 = (1+nu_v)/(2*(-7+5*nu_v))
    C2 = 15*(-2+nu_v)/(4*(-7+5*nu_v))
    source_coords = (lat, lon)

    results = []

    for data in xdata:

        dim = data[0]
        station_coords = (float(data[1]),float(data[2]))
        # print(station_coords)

        y_distance = mo.wgs84_distance( source_coords, (station_coords[0], source_coords[1]) )
        x_distance = mo.wgs84_distance( source_coords, (source_coords[0], station_coords[1]) )
        x_distance = x_distance * np.sign(station_coords[1] - source_coords[1])
        y_distance = y_distance * np.sign(station_coords[0] - source_coords[0])

        R3 = (x_distance**2 + y_distance**2 + source_depth**2)**(3/2)
        result = None
        if dim == 'x':
            result = amplitude *alpha_rad**3*(1+(alpha_rad/source_depth)**3*(C1+C2*source_depth**2/R3**(2/3))) * x_distance / R3
        elif dim == 'y':
            result = amplitude *alpha_rad**3*(1+(alpha_rad/source_depth)**3*(C1+C2*source_depth**2/R3**(2/3))) * y_distance / R3
        elif dim == 'z':
            result = amplitude *alpha_rad**3*(1+(alpha_rad/source_depth)**3*(C1+C2*source_depth**2/R3**(2/3))) * source_depth / R3
        else:
            print("Did not understand dimension")

        results.append(result)
    return results


def closed_pipe(xdata, lat, lon, source_depth, amplitude, pipe_delta):
    '''
    Compute the surface deformation due to changes in a closed pipe source

    For reference, see "Volcano Deformation", Dzurisin 2006, pg 292
    (http://link.springer.com/book/10.1007/978-3-540-49302-0)

    @param xdata: List of the position data with each array element containing [ direction (x, y, or z), lat, lon ]
    @param lat: Latitude of source
    @param lon: Longitude of source
    @param source_depth: Depth of source
    @param amplitude: Ampltiude of source
    @param pipe_delta: Pipe delta from source depth to top/bottom

    @return list of resulting deformation for each point in xdata
    '''
    nu_v = .25
    source_coords = (lat, lon)

    results = []

    for data in xdata:

        dim = data[0]
        station_coords = (float(data[1]),float(data[2]))
        # print(station_coords)

        y_distance = mo.wgs84_distance( source_coords, (station_coords[0], source_coords[1]) )
        x_distance = mo.wgs84_distance( source_coords, (source_coords[0], station_coords[1]) )
        x_distance = x_distance * np.sign(station_coords[1] - source_coords[1])
        y_distance = y_distance * np.sign(station_coords[0] - source_coords[0])

        result = None
        c1 = source_depth - pipe_delta
        c2 = source_depth + pipe_delta
        R_1 = (x_distance**2 + y_distance**2 + c1**2)**(1/2)
        R_2 = (x_distance**2 + y_distance**2 + c2**2)**(1/2)
        r2  = (x_distance**2 + y_distance**2)
        if dim == 'x':
            result = amplitude *((c1/R_1)**3+2*c1*(-3+5*nu_v)/R_1+(5*c2**3*(1-2*nu_v)-2*c2*r2*(-3+5*nu_v))/R_2**3) * x_distance / r2
        elif dim == 'y':
            result = amplitude *((c1/R_1)**3+2*c1*(-3+5*nu_v)/R_1+(5*c2**3*(1-2*nu_v)-2*c2*r2*(-3+5*nu_v))/R_2**3) * y_distance / r2
        elif dim == 'z':
            result = - amplitude *(c1**2/R_1**3+2*(-2+5*nu_v)/R_1+(c2**2*(3-10*nu_v)-2*r2*(-2+5*nu_v))/R_2**3)
        else:
            print("Did not understand dimension")

        results.append(result)
    return results


def constant_open_pipe(xdata, lat, lon, source_depth, amplitude, pipe_delta):
    '''
    Compute the surface deformation due to changes in a constant width open pipe source

    For reference, see "Volcano Deformation", Dzurisin 2006, pg 295
    (http://link.springer.com/book/10.1007/978-3-540-49302-0)

    @param xdata: List of the position data with each array element containing [ direction (x, y, or z), lat, lon ]
    @param lat: Latitude of source
    @param lon: Longitude of source
    @param source_depth: Depth of source
    @param amplitude: Ampltiude of source
    @param pipe_delta: Pipe delta from source depth to top/bottom

    @return list of resulting deformation for each point in xdata
    '''
    nu_v = .25
    source_coords = (lat, lon)

    results = []

    for data in xdata:

        dim = data[0]
        station_coords = (float(data[1]),float(data[2]))
        # print(station_coords)

        y_distance = mo.wgs84_distance( source_coords, (station_coords[0], source_coords[1]) )
        x_distance = mo.wgs84_distance( source_coords, (source_coords[0], station_coords[1]) )
        x_distance = x_distance * np.sign(station_coords[1] - source_coords[1])
        y_distance = y_distance * np.sign(station_coords[0] - source_coords[0])

        result = None
        c1 = source_depth + pipe_delta
        c2 = source_depth - pipe_delta
        R_1 = (x_distance**2 + y_distance**2 + c1**2)**(1/2)
        R_2 = (x_distance**2 + y_distance**2 + c2**2)**(1/2)
        r2  = (x_distance**2 + y_distance**2)
        if dim == 'x':
            result = amplitude *((c1/R_1)**3-2*c1*(1+nu_v)/R_1+(c2**3*(1+2*nu_v)+2*c2*r2*(1+nu_v))/R_2**3)* x_distance / r2
        elif dim == 'y':
            result = amplitude *((c1/R_1)**3-2*c1*(1+nu_v)/R_1+(c2**3*(1+2*nu_v)+2*c2*r2*(1+nu_v))/R_2**3)* y_distance / r2
        elif dim == 'z':
            result = - amplitude *(c1**2/R_1**3-2*nu_v/R_1+(-c2**2+2*R_2**2*nu_v)/R_2**3)
        else:
            print("Did not understand dimension")

        results.append(result)
    return results


def rising_open_pipe(xdata, lat, lon, source_depth, amplitude, pipe_delta,open_pipe_top):
    '''
    Compute the surface deformation due to changes in a rising width amplitude open pipe source

    For reference, see "Volcano Deformation", Dzurisin 2006, pg 295
    (http://link.springer.com/book/10.1007/978-3-540-49302-0)

    @param xdata: List of the position data with each array element containing [ direction (x, y, or z), lat, lon ]
    @param lat: Latitude of source
    @param lon: Longitude of source
    @param source_depth: Depth of source
    @param amplitude: Ampltiude of source
    @param pipe_delta: Pipe delta from source depth to top/bottom
    @param open_pipe_top: Depth of the top of the open pipe

    @return list of resulting deformation for each point in xdata
    '''
    nu_v = .25
    source_coords = (lat, lon)

    results = []

    for data in xdata:

        dim = data[0]
        station_coords = (float(data[1]),float(data[2]))
        # print(station_coords)

        y_distance = mo.wgs84_distance( source_coords, (station_coords[0], source_coords[1]) )
        x_distance = mo.wgs84_distance( source_coords, (source_coords[0], station_coords[1]) )
        x_distance = x_distance * np.sign(station_coords[1] - source_coords[1])
        y_distance = y_distance * np.sign(station_coords[0] - source_coords[0])

        result = None
        c0 = open_pipe_top
        c1 = source_depth + pipe_delta
        R_0 = (x_distance**2 + y_distance**2 + c0**2)**(1/2)
        R_1 = (x_distance**2 + y_distance**2 + c1**2)**(1/2)
        r2  = (x_distance**2 + y_distance**2)
        if dim == 'x':
            result = amplitude *(-(c0**2/R_0**3)+2*nu_v/R_0+(c1**2-2*(c1**2+r2)*nu_v)/R_1**3)* x_distance / c1
        elif dim == 'y':
            result = amplitude *(-(c0**2/R_0**3)+2*nu_v/R_0+(c1**2-2*(c1**2+r2)*nu_v)/R_1**3)* y_distance / c1
        elif dim == 'z':
            result = -amplitude *((c0**3/R_0**3)-c1**3/R_1**3+c1*(-1+2*nu_v)/R_1+c0*(1-2*nu_v)/R_0+(-1+2*nu_v)*np.log(c0+R_0)-(-1+2*nu_v)*np.log(c1+R_1))/ c1
        else:
            print("Did not understand dimension")

        results.append(result)
    return results


def sill(xdata, lat, lon, source_depth, amplitude):
    '''
    Compute the surface deformation due to changes in a sill-like source

    For reference, see "Volcano Deformation", Dzurisin 2006, pg 297
    (http://link.springer.com/book/10.1007/978-3-540-49302-0)

    @param xdata: List of the position data with each array element containing [ direction (x, y, or z), lat, lon ]
    @param lat: Latitude of source
    @param lon: Longitude of source
    @param source_depth: Depth of source
    @param amplitude: Ampltiude of source

    @return list of resulting deformation for each point in xdata
    '''
    source_coords = (lat, lon)
    results = []

    for data in xdata:

        dim = data[0]
        station_coords = (float(data[1]),float(data[2]))
        # print(station_coords)

        y_distance = mo.wgs84_distance( source_coords, (station_coords[0], source_coords[1]) )
        x_distance = mo.wgs84_distance( source_coords, (source_coords[0], station_coords[1]) )
        x_distance = x_distance * np.sign(station_coords[1] - source_coords[1])
        y_distance = y_distance * np.sign(station_coords[0] - source_coords[0])

        R5 = (x_distance**2 + y_distance**2 + source_depth**2)**(5/2)
        result = None
        if dim == 'x':
            result = amplitude * x_distance * source_depth**2 / R5
        elif dim == 'y':
            result = amplitude * y_distance * source_depth**2 / R5
        elif dim == 'z':
            result = amplitude * source_depth**3 / R5
        else:
            print("Did not understand dimension")

        results.append(result)
    return results


def dirEigenvectors(coord_list, pca_comps,pdir='H'):
    '''
    Takes eigenvectors (north and east) and forces them to point "outward"

    Flips the sign of the projection if needed so that eigenvectors point
    outward. Needed because the "positive" direction for PCA is arbitrary

    @param coord_list: Location of stations for projecting the eigenvectors
    @param pca_comps: PCA components
    @param pdir: PCA direction, vertical or horizontal

    @return station_lat_list: the station latitude coordinates
    @return station_lon_list: the station longitude coordinates
    @return ev_lat_list: the properly origented corresponding eigenvector latitude component
    @return ev_lon_list: the properly origented corresponding eigenvector longitude component
    @return direction scale factor (1 for no flip, or -1 for flip)
    '''
    station_lat_list = []
    station_lon_list = []
    ev_lat_list = []
    ev_lon_list = []
    if pdir=='H':
        for i in range(int(len(pca_comps)/2)):
            station_lat_list.append(coord_list[i][0])
            station_lon_list.append(coord_list[i][1])
            ev_lat_list.append(pca_comps[2*i])
            ev_lon_list.append(pca_comps[2*i+1])

    elif pdir=='V':
        for i in range(int(len(pca_comps))):
            station_lat_list.append(coord_list[i][0])
            station_lon_list.append(coord_list[i][1])
            ev_lat_list.append(pca_comps[i])
            ev_lon_list.append(0)

    station_lat_list = np.array(station_lat_list)
    station_lon_list = np.array(station_lon_list)
    ev_lat_list = np.array(ev_lat_list)
    ev_lon_list = np.array(ev_lon_list)

    if pdir=='H':
        # mean coordinate
        mean_lon = np.mean(station_lon_list)
        mean_lat = np.mean(station_lat_list)

        # project the eigen vector along the direction relative to mean
        mean_angle = np.arctan2(station_lat_list-mean_lat,station_lon_list-mean_lon)
        eigv_angle = np.arctan2(ev_lat_list,ev_lon_list)
        proj_angle = np.abs(mean_angle - eigv_angle)

        if np.sum(proj_angle<np.pi/2)<np.sum(proj_angle>=np.pi/2):
            return station_lat_list, station_lon_list, ev_lat_list*-1, ev_lon_list*-1, -1
        else:
            return station_lat_list, station_lon_list, ev_lat_list, ev_lon_list, 1
    else:
        return station_lat_list, station_lon_list, ev_lat_list, ev_lon_list, 1


def datetimeToNumber(in_time):
    '''
    Converts input pandas Timestamp or pandas DatetimeIndex to unix time

    @param in_time: Input pandas timestamp or pandas DatetimeIndex

    @return unix time
    '''
    if isinstance (in_time, pd.Timestamp):
        return time.mktime(in_time.timetuple())
    elif isinstance(in_time, pd.DatetimeIndex):
        return in_time.map(lambda t: time.mktime(t.timetuple()))
    else:
        return in_time



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from skdiscovery.data_structure.framework import PipelineItem
from scipy import signal


class LowPassFilter(PipelineItem):
    '''
    A remez low pass filter for table data
    '''
    
    def __init__(self, str_description, ap_paramList):
        '''
        Initialize LowPassFilter

        @param str_description: String describing filter
        @param ap_paramList[ntaps]: number of filter taps
        @param ap_paramList[fpassf_per]: frequency passband ratio/percentage
        @param ap_paramList[fstopf_per]: frequency stopband ratio/percentage
        @param ap_paramList[wghts]: band importance weights
        @param ap_paramList[miter]: maximum number of iterations for generating the filter
        '''
        
        super(LowPassFilter, self).__init__(str_description, ap_paramList)
        self.ap_paramNames = ['ntaps','fpassf_per','fstopf_per','weights','miter']
        
    def process(self, obj_data):
        ''' 
        Apply lowpass filter to data set
        
        @param obj_data: Input data. Changes are made in place.
        '''

        column_names = obj_data.getDefaultColumns()
        
        ntaps = self.ap_paramList[0]()
        fpassf_per = self.ap_paramList[1]()
        fstopf_per = self.ap_paramList[2]()
        wghts = self.ap_paramList[3]()
        miter = self.ap_paramList[4]()
        b_filt=signal.fir_filter_design.remez(numtaps=ntaps,
                                              bands=[0,fpassf_per,fstopf_per,.5],
                                              desired=[1,0],weight=wghts,maxiter=miter)
        
        for label, data in obj_data.getIterator():
            for column in column_names:
                obj_data.updateData(label, data.index, column, signal.filtfilt(b_filt,1,data))
                    



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Evan Wojciechowski
# This software is part of projects sponsored by NASA and NSF (PI: V. Pankratius)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# written in python 3.6 by evan wojciechowski
# miscellaneous utilities for use in various time-series decomposition functions
# kept here in an attempt to reduce clutter
#! rename this (and then fix imports in dependencies)

import os, sys, warnings

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


# immutable stuff
types = [int, float, complex, np.float32, np.float64, np.int32, np.int64, np.complex64, np.complex128,]
coldict = {0 : 'C0', 1 : 'C1', 2 : 'C2', 3 : 'C3', 4 : 'C4', 5 : 'C5', 6 : 'C6', 7 : 'C7', 8 : '#92C7Ed', 9 : '#FFBB80', 10 : '#9BE49B', 11 : '#EB9393', 12 : '#C0A6D8', 13 : '#D2B3AC', 14 : '#E995D0', 15 : '#BFBFBF'}

# helper functions

def lin_trend(inData, toReturn = 'eval'):
    real = np.isfinite(inData)
    rng  = np.arange(len(inData))

    lin_fit = np.poly1d(np.polyfit(rng[real], inData[real], 1))(rng[real])

    return lin_fit


def index_scale(toScale, endRange = []):
    """
    quick linear adjustment to some input index
    primary: wavelet transforms (pywt)
    @param toScale: input index array-like to be scaled
    @param endRange: optional input list to have (at least) two values that bound output index range; assumes min is [0], max is [1]
    """

    if len(endRange) is 0:
        endRange = [0, len(toScale)]

    iMin = np.min(toScale)
    iMax = np.max(toScale)

    if (iMax - iMin) is not 0:
        delta = iMax - iMin
    else:
        delta = 1 / iMax

    return np.array(((toScale - iMin) / delta) * (endRange[1] - endRange[0]) + endRange[0])

def block_output():
    """
    blocks print output from functions and suppresses warnings
        mostly used to make output look nice, ignoring (often known) minor errors
    primary: emd (pyhht)
    """

    sys.stdout = open(os.devnull, 'w')
    warnings.filterwarnings('ignore')

def enable_output():
    """
    re-enables output, foil of blockOutput
    primary: emd (pyhht)
    """
    sys.stdout = sys.__stdout__
    warnings.filterwarnings('default')


def mod_data(inData, inIndex = None, makeType = None,):
    """
    modifies data for run_spiral so that plotted data is uniform
    @param inData: data values to be used as intensity
    @param inIndex: data values to be used as radial and angular components (once period is applied)
    @param makeType: variable tied to makeIndex which will create different types of index if necessary
    """

    makeIndex = False

    # if a separate index is provided
    if inIndex is not None:
        if len(inIndex) != len(inData):
            # sanity checks on index length
            if len(inIndex) > len(inData):
                # uses a matching amount of indices
                outIndex = inIndex[:len(inData)]
            else:
                # not enough indices, must make some new index
                makeIndex = True

    # data checking; np.ndarray and pd.core.series.Series should be the two most common types of data structure
    if isinstance(inData, np.ndarray):
        # no chance for index, only contains amplitude
        outData   = inData
        makeIndex = True 

    elif isinstance(inData, pd.core.series.Series):
        inIndex = inData.index
        #! need to check index type
        dtype = inIndex.dtype

        if dtype in types:
            # no conversion
            #! still may need to do something
            pass

        elif dtype == 'datetime64[ns]':
            # converts to np.float64, starts at zero (linear removal), unit is days
            inIndex  = inIndex.to_julian_date()
            inIndex -= inIndex.min()

        else:
            print('format not accounted for (FIX)')
            makeIndex = True

        outData  = inData.values.astype('float64')
        outIndex = inIndex

    else:
        print('format unaccounted for (FIX)')
        return

    if inIndex is None:
        # final check to see if any index was assigned
        makeIndex = True

    # creates an index if needed
    #! very rough sketch
    if makeIndex:
        if (makeType == int) or (makeType is None):
            outIndex = np.arange(len(inData))
        elif makeType.lower() == 'fourier':
            outIndex = np.linspace(0, 2*np.pi, len(inData))

    # eliminates potential NaN values from plotted data
    real     = np.isfinite(outData)
    outData  = outData[real]
    outIndex = outIndex[real]

    # returns final product for plotting 
    return outData, outIndex




# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from skdiscovery.data_structure.framework import PipelineItem
from skdiscovery.utilities.patterns import pbo_tools
from scipy.optimize import curve_fit
from collections import defaultdict
import pandas as pd
import numpy as np


class HTanFilter(PipelineItem):
    '''
    Filter to subtract arctan fit from data

    [DEPRECATED] [will be removed]
    '''
    def __init__(self, str_description, t0, amplitude=5, timescale=1., offset=0,
                 slope=0, labels=None, column_names=None, start_time_limit = None,
                 end_time_limit = None, start = None, end = None):
        '''
        Fit and remove hyperbolic tangent function from data.

        @param str_description: String description of data
        @param t0: Initial time offset of arctangent
        @param amplitude: Initial amplitude of arctangent
        @param timescale: Timescale of fit
        @param offset: Initial Y offset of arctangent
        @param slope: Slope of the data 
        @param labels: Labels to apply arctangent function to
        @param column_names: Column names to apply arctanget function to
        @param start_time_limit: Starting time bound for fit to arctan (default: no bound)
        @param end_time_limit: Ending time bound for fit to arctan (default: no bound)
        @param start: Index of the first data point to fit (default: index of first data point)
        @param end: Index of the last data point to fit (default: index of last data point)
        '''
        self.a = amplitude
        self.t0 = t0
        self.c = timescale
        self.slope = slope
        self.offset = offset
        self.labels = labels
        self.column_names = column_names
        self.start_time_limit = start_time_limit
        self.end_time_limit = end_time_limit
        self.start = start
        self.end = end

        super(HTanFilter, self).__init__(str_description, [])


    def process(self, obj_data):
        ''' 
        Apply Arctangent filter to data param.

        @param obj_data: Input data. Changes are made in place.
        '''
        
        parameters = [self.a, pbo_tools.datetimeToNumber(self.t0), self.c, self.offset, self.slope]

        def fitfunc(t,a,t0,c,b,m):

            # return np.piecewise(t, [t <= t0, t > t0], [ lambda t: a*np.arctan((t-t0)/c)+b1+m1*t,
            #                                             lambda t: a*np.arctan((t-t0)/c)+b2+m2*t ] )

            return a*np.tanh((t-t0)/c)+b+m*t

        labels = self.labels
        column_names = self.column_names

        results = defaultdict(pd.DataFrame)


        if self.start_time_limit is not None:
            start_time_limit = pbo_tools.datetimeToNumber(self.start_time_limit)
        else:
            start_time_limit = -np.inf

        if self.end_time_limit is not None:
            end_time_limit = pbo_tools.datetimeToNumber(self.end_time_limit)
        else:
            end_time_limit = np.inf


        bounds = ( (-np.inf, start_time_limit, -np.inf, -np.inf, -np.inf ),
                   ( np.inf, end_time_limit,    np.inf,  np.inf,  np.inf) )



        for label, data, err in obj_data.getIterator():

            if (labels is None or label in labels) and \
               (column_names is None or data.name in column_names):


                index = data.index

                if self.start is None:
                    start = index[0]
                else:
                    start = self.start

                if self.end is None:
                    end = index[-1]
                else:
                    end = self.end

                # new_index = pbo_tools.datetimeToNumber(data.index)
                new_index = pbo_tools.datetimeToNumber(data.loc[start:end].index)
                full_new_index = pbo_tools.datetimeToNumber(data.index)
                
                
                pA, success = curve_fit(fitfunc, new_index, data.loc[start:end], p0=parameters, sigma = err.loc[start:end],
                                        absolute_sigma=True, bounds=bounds)

                data.update(data - fitfunc(full_new_index, *pA))

                results[label][data.name] = pd.Series(fitfunc(full_new_index, *pA), index=index)

        obj_data.addResult(self.str_description, pd.Panel(results))



'''
This is a modified version of rforward.py, which can be found at:
https://github.com/paramiko/paramiko/blob/25dd096da065b1bc2f35c1a62d8a7055b022818b/demos/rforward.py

This version removes the command line interface, global variables, and creates tunnels in a non blocking thread.
An additional class has been added to facilitate the creation of tunnels

The original version is Copyright (C) 2008 Robey Pointer <robeypointer@gmail.com>

Since the original version is LGPL v2.1 or later, this is also LGPL v2.1 or later

Changes made to original

2017-03-30: Converted example into library for creating reverse ssh tunnels

2017-04-11: Updates to verbosity option and reverse tunneling

2017-07-10: Updates to documentation
'''


import socket
import threading
import select
import paramiko

def print_verbose(s, verbose=False):
    '''
    Print statement if verbose is True

    @param s: Statement to print
    @param verbose: Print only if verbose is True
    '''
    if verbose:
        print(s)


def handler(chan, host, port, verbose=False):
    '''
    Handler is responsible for sending and receiving data through ssh tunnel

    @param chan: SSH Channel for transferring data
    @param host: Address of remote host 
    @param port: Port to forward
    @param verbose: Print status information
    '''
    sock = socket.socket()
    try:
        sock.connect((host, port))
    except Exception as e:
        print_verbose('Forwarding request to %s:%d failed: %r' % (host, port, e), verbose)
        return
    
    print_verbose('Connected!  Tunnel open %r -> %r -> %r' % (chan.origin_addr,
                                                        chan.getpeername(), (host, port)), verbose)
    while True:
        r, w, x = select.select([sock, chan], [], [])
        if sock in r:
            data = sock.recv(1024)
            if len(data) == 0:
                break
            chan.send(data)
        if chan in r:
            data = chan.recv(1024)
            if len(data) == 0:
                break
            sock.send(data)
            
    chan.close()
    sock.close()
    print_verbose('Tunnel closed from %r' % (chan.origin_addr,), verbose)


def reverse_forward_tunnel(server_port, remote_host, remote_port, transport, check=30, verbose=False):
    '''
    Creates a reverse ssh tunnel

    @param server_port: Port on local host
    @param remote_host: Address of remote host
    @param remote_port: Port of remote host
    @param transport: SSH Transport
    @param check: Amount of time to wait in seconds when opening up a channel
    @param verbose: Print status information

    @return Thread running reverse ssh tunnel, event used to close ssh tunnel, 
            list of child threads started by main thread
    '''

    transport.request_port_forward('', server_port)
    
    event = threading.Event()
    child_threads = []
    
    def accept_tunnels(event):
        '''
        This function spawns new connections as they are needed
        
        @param event: When this event is set, this function will complete
        '''
        while not event.is_set():
            chan = transport.accept(check)
            if chan is None:
                continue
            thr = threading.Thread(target=handler, args=(chan, remote_host, remote_port, verbose))
            thr.setDaemon(True)
            thr.start()
            child_threads.append(thr)
                
    accept_thread = threading.Thread(target=accept_tunnels,args=[event])
    accept_thread.setDaemon(True)
    accept_thread.start()
    
    return accept_thread, event, child_threads
            

class ReverseTunnel(object):
    '''
    Create a reverse ssh tunnel
    '''

    def __init__(self, server_address, username, key_filename, server_port,
                 remote_host, remote_port, check=30, verbose=False):
        '''
        Initialize ReverseTunnel object

        @param server_address: Local server address
        @param username: Valid username on remote host
        @param key_filename: Filename of ssh key associated with remote host
        @param server_port: Local port
        @param remote_host: Address of remote host
        @param remote_port: Remote port
        @param check: Amount of time to wait in seconds when opening up a channel
        @param verbose: Print status information
        '''

        self.server_address = server_address
        self.username       = username
        self.key_filename   = key_filename
        self.server_port    = server_port
        self.remote_host    = remote_host
        self.remote_port    = remote_port
        self.check          = check
        self.verbose        = verbose

        self.ssh = None
        self.event = None

    
    def create_reverse_tunnel(self):
        '''
        Create the reverse tunnel
        '''
        self.ssh = paramiko.SSHClient()
        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.ssh.connect(self.server_address, username=self.username, key_filename=self.key_filename)
        self.running_thread, self.event, self.child_threads = reverse_forward_tunnel(self.server_port, self.remote_host,
                                                                                     self.remote_port, self.ssh.get_transport(),
                                                                                     self.check)


    def __del__(self):
        '''
        Deconstructor
        '''
        if self.ssh != None:
            self.ssh.close()

        if self.event != None:
            self.event.set()



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from skdiscovery.data_structure.framework import PipelineItem
from skdaccess.utilities.pbo_util import removeAntennaOffset


class AntennaOffset(PipelineItem):
    '''
    Applies corrections to fix offsets in PBO GPS data induced by antenna changes
    '''

    def __init__(self, str_description, antenna_data, min_diff = 0.0, column_list = None):
        '''
        Initialize AntennaOffset function

        @param str_description: String describing the filter
        @param antenna_data: Data containing the log of antenna changes
        @param min_diff: Difference in position needed to be considered an offset
        @param column_list: Names of the columns to apply the function to
        '''

        super(AntennaOffset, self).__init__(str_description, [])
        self.antenna_data = antenna_data
        self.column_list = column_list
        self.min_diff = min_diff


    def process(self, obj_data):
        '''
        Applies the function to the data, updating in place

        @param obj_data: Table data wrapper
        '''
        
        if self.column_list == None:
            column_names = obj_data.getDefaultColumns()

        else:
            column_names = self.column_list


        for label, data in obj_data.getIterator():
            for column in column_names:
                try:
                    updated_data = removeAntennaOffset(self.antenna_data[label], data[column], min_diff = self.min_diff)
                    index = data[column].index
                    obj_data.updateData(label,index,column,updated_data)
                    
                except KeyError:
                    pass
                    



from .dataremover import DataRemover
from .hyperbolictan import HTanFilter
from .interpolate import InterpolateFilter
from .kalman import KalmanFilter
from .lowpass import LowPassFilter
from .table_filter import TableFilter
from .median import MedianFilter
from .trend import TrendFilter
from .offset_detrend import OffsetDetrend
from .snow_remover import SnowRemover
from .weighted_average import WeightedAverage
from .propagate_nans import PropagateNaNs
from .antenna_offset import AntennaOffset
from .stabilization import StabilizationFilter
from .geolocation import GeoLocationFilter
from .combine_columns import CombineColumns
from .calibrate_grace import CalibrateGRACE
from .resample import Resample
from .normalize import NormalizeFilter

__all__ = ['DataRemover', 'HTanFilter', 'InterpolateFilter',
           'KalmanFilter', 'LowPassFilter', 'MedianFilter',
           'TrendFilter','OffsetDetrend', 'GeolocationFilter',
           'StabilizationFilter']



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Evan Wojciechowski
# This software is part of projects sponsored by NASA and NSF (PI: V. Pankratius)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from .vis_utils import *

import matplotlib.pyplot as plt
import numpy as np

def calc_DFT(t, y):
    '''
    Calculates discrete Fourier transform using np.fft.fft

    @param t: Time array
    @param y: Y (data amplitude) array

    @return Tuple of post-FT frequencies and coefficients
    '''

    tf = np.linspace(t[0], 1/(2*t[-1]/len(t)), len(t)/2)
    yf = np.fft.fft(y)

    return tf, yf

def plot_DFT(tIndex, yData, collage = False, show = True, suptitle = '', hori = True):
    '''
    Plots input data and Fourier transformed coefficients in a subplot grid

    @param tIndex: Input time index for series
    @param yData: Input data amplitude
    @param collage: Boolean that can optionally return certain plot parameters for external usage
    @param show: Boolean to show plot immediately after plot creation   
    @param suptitle: Optional string to add as a plot title
    @param hori: Boolean that optionally changes the orientation of the subplot configuration
    '''

    tfIndex, yfData = calc_DFT(tIndex, yData)

    plotArgs = tIndex, yData, tfIndex, (2/len(tIndex))*np.abs(yfData[:len(tIndex)//2])

    if collage:
        return plotArgs # allows use in some other function

    if hori:
        fig, axes = plt.subplots(1, 2, figsize = (12, 6.75))
    else:
        fig, axes = plt.subplots(2, 1, figsize = (12, 6.75))
    
    if suptitle != '':
        fig.suptitle(suptitle, fontsize = 18)

    axes[0].plot(plotArgs[0], plotArgs[1], color = '#0066FF')
    axes[1].plot(plotArgs[2], plotArgs[3], color = '#FF6600')

    axes[0].set_title('Time series', fontsize = 18)
    axes[0].set_xlabel('Data Index', fontsize = 16)
    axes[0].set_ylabel('Data Amplitude', fontsize = 16)
    axes[1].set_title('Fourier spectrum', fontsize = 16)
    axes[1].set_xlabel('Frequency', fontsize = 16)
    axes[1].set_ylabel('Frequency Amplitude', fontsize = 16)

    for ax in axes:
        ax.tick_params(axis = 'both', labelsize = 12)

    if show:
        plt.show()

#%# wrappers #%#
def run_plotDFT(inData, inIndex = None, collage = False, show = True, suptitle = '', hori = True):
    '''
    Wrapper for plot_DFT

    @param inData: Input data for plotting
    @param inIndex: Possible input index to use in calculating DFT
    @param collage: Boolean that can optionally return certain plot parameters for external usage
    @param show: Boolean to show plot immediately after plot creation   
    @param suptitle: Optional string to add as a plot title
    @param hori: Boolean that optionally changes the orientation of the subplot configuration
    '''


    y, t = mod_data(inData, inIndex, makeType = 'fourier')
    plot_DFT(t, y, collage = collage, show = show, suptitle = suptitle, hori = hori)




# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from skdiscovery.data_structure.framework import PipelineItem
import numpy as np
import matplotlib.pyplot as plt
import math

class Plotter(PipelineItem):
    '''
    Make a plot of series data
    '''
    def __init__(self, str_description, num_columns = 3, errorbars=False, width=13, height=4, **kwargs):
        '''
        Initialize Plotter
        
        @param str_description: String describing accumulator
        @param num_columns: Number of columns to use when plotting data
        @param errorbars: Flag indicating if errorbars should be used
        @param width: Total width of all columns combined
        @param height: Height of single row of plots
        @param **kwargs: Any additional keyword arguments are passed on to matplotlib
        '''
        self.kwargs = kwargs
        self.num_columns = num_columns
        self.errorbars = errorbars
        self.height = height
        self.width = width
        super(Plotter, self).__init__(str_description, [])
    
    def process(self, obj_data):
        '''
        Plot each column in obj_data

        @param obj_data: Data Wrapper
        '''

        width = self.width
        height = self.height
        
        labels = []
        data = []
        errors = []
        for label, series, err in obj_data.getIterator():
            data.append(series)
            errors.append(err)
            labels.append(label)

        # if len(data) > self.num_columns:
        #     width *= self.num_columns
        # else:
        #     width *= len(data)

        if len(labels) > 0:

            rows = math.ceil(len(labels) / self.num_columns)
            height *= rows

            figure = plt.figure()
            figure.set_size_inches(width, height, True)


            for label, series, err, num in zip(labels, data, errors, range(1,len(labels)+1)):
                plt.subplot(rows, self.num_columns, num)
                plt.title(series.name)
                plt.ylabel(label)

                plt.xticks(rotation=45)

                if self.errorbars:
                    plt.errorbar(np.array(series.index),np.array(series), yerr=np.array(err), **self.kwargs)
                else:
                    plt.plot(series, **self.kwargs)

            plt.tight_layout()

            if(obj_data.run_id > -1):
                figure.suptitle( "Run: " + str(obj_data.run_id), y=1.02)




# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and
# NASA AIST Project "Computer-Aided Discovery of Earth Surface
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from skdiscovery.data_structure.framework.base import PipelineItem
import numpy as np


class WeightedAverage(PipelineItem):
    ''' This filter performs a rolling weighted average using standard deviations as weight '''

    def __init__(self, str_description, ap_paramList, column_names, std_dev_column_names=None, propagate_uncertainties=False):
        '''
        Initializes a WeightedAverage object

        @param str_description: String describing filter
        @param ap_paramList[window]: Window to use for computing rolling weighted average
        @param column_names: Names of columns to apply the weighted average
        @param std_dev_column_names: Names of columns of the standard deviations. If none a regular mean is computed.
        @param propogate_unceratinies: Propogate uncertainties assuming uncorrelated errors
        '''

        super(WeightedAverage,self).__init__(str_description,ap_paramList)
        self.column_names = column_names
        self.std_dev_column_names = std_dev_column_names
        self.propagate_uncertainties = propagate_uncertainties


    def process(self, obj_data):
        '''
        Apply the moving (weighted) average filter to a table data wrapper.n

        Changes are made in place.

        @param obj_data: Input table data wrapper
        '''

        window = self.ap_paramList[0]()

        for label, data in obj_data.getIterator():

            if self.std_dev_column_names != None:
                for column, std_dev_column in zip(self.column_names,
                                                  self.std_dev_column_names):

                    weights = 1 / data[std_dev_column]**2
                    weighted_data = data[column] * weights

                    scale = weights.rolling(window=window,center=True, min_periods=1).sum()

                    weighted_average = weighted_data.rolling(window=window, center=True, min_periods=1).sum() / scale

                    obj_data.updateData(label, weighted_average.index, column,weighted_average)

                    if self.propagate_uncertainties:
                        # Uncertainty determined using the standard error propagation technique
                        # Assumes data is uncorrelated
                        uncertainty =  1 / np.sqrt(scale)
                        obj_data.updateData(label, uncertainty.index, std_dev_column, uncertainty)


            else:
                for column in self.column_names:

                    weighted_average = data[column].rolling(window=window, center=True, min_periods=1).mean()

                    obj_data.updateData(label,weighted_average.index,column,weighted_average)



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from collections import defaultdict

from skdiscovery.data_structure.framework import PipelineItem
from scipy.stats import skew
import numpy as np

class Skew(PipelineItem):
    ''' Calculates the skew of table data '''
    def process(self, obj_data):
        '''
        Apply Skew analysis with results added to the data wrapper

        @param obj_data: Data wrapper
        '''

        column_names = obj_data.getDefaultColumns()
        
        results = defaultdict(dict)
        # for label, frame in tqdm(obj_data.getIterator()):
        for label, frame in obj_data.getIterator():
            for column in column_names:
                # dropping missing data in order to remove top and bottom 2%
                data = frame[column].dropna()
                # Remove top and bottom 2%
                rem_num = round(len(data)*0.02)
                res = skew(data.sort_values(ascending=True)[rem_num:-rem_num])
                if isinstance(res, np.ma.masked_array):
                    res = np.float(res.data)
                results[label][column] = res
                obj_data.addResult(self.str_description, results)



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from skdiscovery.data_structure.framework.base import PipelineItem
from skdiscovery.data_structure.framework import DiscoveryPipeline
from skdiscovery.data_structure.generic.accumulators import DataAccumulator
from skdiscovery.data_structure.table.filters import CalibrateGRACE, Resample
from skdiscovery.data_structure.framework.stagecontainers import *
from skdaccess.framework.param_class import *
from skdaccess.geo.grace import DataFetcher as GDF
from skdaccess.geo.gldas import DataFetcher as GLDASDF

import numpy as np


class GraceFusion(PipelineItem):
    ''' 
    Fuses GRACE equivelent water depth time series

    Works on table data (original data from http://grace.jpl.nasa.gov/data/get-data/monthly-mass-grids-land/) 
    '''

    def __init__(self, str_description, metadata, column_data_name = 'Grace', column_error_name = 'Grace_Uncertainty', gldas = "Off"):
        '''
        Initialize Grace Fusion item

        @param str_description: String describing item
        @param metadata: Metadata that contains lat,lon coordinates based on data labels
        @param column_data_name: Name of column for GRACE data
        @param column_error_name: Grace Uncertainty column name
        @param gldas: Indicating use of the global land data assimilation water model
        '''
        super(GraceFusion, self).__init__(str_description, [])
        self.metadata = metadata.copy()
        self.column_data_name = column_data_name
        self.column_error_name = column_error_name
        # remove_sm_and_snow
        self.gldas = gldas
        self._tileCache = None

        
    def process(self, obj_data):
        ''' 
        Adds columns for GRACE data and uncertainties

        @param obj_data: Input DataWrapper, will be modified in place
        '''

        # Only perform fusion if data exists
        if obj_data.getLength() > 0:
            # if nothing is cached, load generate and cache data
            if self._tileCache == None:
                self._tileCache = self._gen_tile_cache(obj_data)

            # grab correct index from the unique locations list to grab the correct GRACE result
            for label, data in obj_data.getIterator():
                try:
                    lat = self.metadata[label]['Lat']
                    lon = self.metadata[label]['Lon']
                except:
                    lat = self.metadata.loc[label,'Lat']
                    lon = self.metadata.loc[label,'Lon']
                    
                idx = (int(np.floor(lat)),int(np.floor(lon)))
                # check if key in cache, and if not overwrite
                if idx not in self._tileCache.keys():
                    self._tileCache = self._gen_tile_cache(obj_data)

                if self.gldas.lower() != 'only':
                    grace_data = self._tileCache[idx]['EWD']
                else:
                    grace_data = self._tileCache[idx]['GLDAS']
                grace_err  = self._tileCache[idx]['EWD_Error']
                obj_data.addColumn(label, self.column_data_name, grace_data)
                obj_data.addColumn(label, self.column_error_name, grace_err)

    def _gen_tile_cache(self,obj_data):
        ''' 
        Builds a data cache of the GRACE results for the unique tiles

        @param obj_data: Input DataWrapper for grabbing data

        @return Dictionary the GRACE results for each coordinate
        '''
        locations = []
        
        start_date = None
        end_date = None
        
        for label, data in obj_data.getIterator():
            try:
                lat = self.metadata[label]['Lat']
                lon = self.metadata[label]['Lon']
            except:
                lat = self.metadata.loc[label,'Lat']
                lon = self.metadata.loc[label,'Lon']
                
            locations.append([lat,lon])

            if start_date == None:
                start_date = data.index[0]
                end_date = data.index[-1]
            else:
                if start_date != data.index[0] \
                or end_date != data.index[-1]:
                    raise RuntimeError("Varying starting and ending dates not supported")
            
        unilocs = [urow for urow in {tuple(row) for row in np.floor(locations)}]

        fetcher_unilocs = [[urow] for urow in {tuple(row) for row in np.floor(locations)}]
        al_locations = AutoListCycle(fetcher_unilocs)
        al_locations_gldas = AutoListCycle(fetcher_unilocs)

        graceDF = GDF([al_locations],start_date, end_date)
        gldasDF = GLDASDF([al_locations_gldas], start_date, end_date)

        

        # fl_interp = InterpolateFilter('Interpolate',[])
        # sc_interp = StageContainer(fl_interp)
        
        # grace_pipe = DiscoveryPipeline(graceDF, [sc_interp,sc_data])

        def getData(datafetcher,length):

            ac_data = DataAccumulator('Data',[])
            sc_data = StageContainer(ac_data)

            fl_grace = CalibrateGRACE('Calibrate')
            sc_grace = StageContainer(fl_grace)

            fl_resample = Resample('Resample',start_date, end_date)
            sc_resample = StageContainer(fl_resample)
            

            pipeline = DiscoveryPipeline(datafetcher, [sc_grace, sc_resample, sc_data])

            pipeline.run(num_cores=1, num_runs = length, perturb_data=True)
            

        
            gpgr = dict()
            for ii in np.arange(len(unilocs)):
                key = str(unilocs[ii][0]) + ', ' + str(unilocs[ii][1])
                gpgr[unilocs[ii]] = pipeline.getResults(ii)['Data'][key]
            return gpgr

        # If we are not removing sm and snow
        # we are down now
        if self.gldas.lower() == 'off':
            grace_data = getData(graceDF, len(unilocs))
            return grace_data

        elif self.gldas.lower() == 'remove':
            # If we are removing sm and snow
            grace_data = getData(graceDF, len(unilocs))            
            gldas_data = getData(gldasDF, len(unilocs))

            for key in grace_data:
                grace = grace_data[key]['Data']
                gldas = gldas_data[key]['Data']

                grace_index = grace.index
                grace.dropna(inplace=True)

                # If no grace data available, no need to remove gldas
                if len(grace) == 0:
                    continue

                # Get matching start and end
                start_grace = grace.index[0]
                end_grace = grace.index[-1]
                start_gldas = gldas.index[0]
                end_gldas = gldas.index[-1]

                start_month = np.max([start_gldas,start_grace]).strftime('%Y-%m')
                end_month = np.min([end_gldas,end_grace]).strftime('%Y-%m')

                # Convert gldas to a data frame
                # and save index
                # gldas = gldas.loc[:,:,'GLDAS'].copy()
                gldas.loc[:,'Date'] = gldas.index

                # Index GLDAS data by month
                new_index = [date.strftime('%Y-%m') for date in gldas.index]
                gldas.loc[:,'Month'] = new_index
                gldas.set_index('Month',inplace=True)

                # select only months that are also in GRACE
                cut_gldas = gldas.loc[[date.strftime('%Y-%m') for date in grace.loc[start_month:end_month,:].index],:]

                # index GLDAS data to GRACE dates
                cut_gldas.loc[:,'Grace Index'] =  grace.loc[start_month:end_month,:].index
                cut_gldas.set_index('Grace Index', inplace=True)


                # Calculate distance between days
                offset_days = cut_gldas.index - cut_gldas.loc[:,'Date']
                offset_days = offset_days.apply(lambda t: t.days)
                cut_gldas.loc[:,'Offset'] = offset_days            


                # Remove any data where the difference between gldas and grace are > 10 days            
                cut_gldas = cut_gldas[np.abs(cut_gldas.loc[:,'Offset']) < 10].copy()

                # Select appropriate Grace Data
                cut_grace = grace.loc[cut_gldas.index,:]

                # Remove contribution of snow + sm to GRACE
                cut_grace.loc[:,'Grace'] = cut_grace.loc[:,'Grace'] - cut_gldas['GLDAS']

                # Now restore to original index, filling in with NaN's
                grace = cut_grace.reindex(grace_index)

                # index, place the result back into the 
                grace_data[key]['Data'] = grace

            # All the snow and sm contribution has been removed,
            # so the dictinoary can now be returned

            return grace_data
    

        elif self.gldas.lower() == 'only':
            return getData(gldasDF, len(unilocs))

        else:
            raise ValueError('Did not understand gldas option: ' + self.gldas.lower())



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import numpy as np
import pandas as pd
import matplotlib
from matplotlib.patches import Polygon
from scipy.spatial import SphericalVoronoi
import pyproj

# utility functions for generating the spherical voronoi tesselation.

def sphericalToXYZ(lat,lon,radius=1):
    ''' 
    Convert spherical coordinates to x,y,z

    @param lat: Latitude, scalar or array
    @param lon: Longitude, scalar or array
    @param radius: Sphere's radius

    @return Numpy array of x,y,z coordinates
    '''
    phi = np.deg2rad(90.0 - lat)
    theta = np.deg2rad(lon % 360)
    x = radius * np.cos(theta)*np.sin(phi)
    y = radius * np.sin(theta)*np.sin(phi)
    z = radius * np.cos(phi)
    
    if np.isscalar(x) == False:
        return np.vstack([x,y,z]).T
    else:
        return np.array([x,y,z])


def xyzToSpherical(x,y,z):
    ''' 
    Convert x,y,z to spherical coordinates

    @param x: Cartesian coordinate x
    @param y: Cartesian coordinate y
    @param z: Cartesian coordinate z

    @return numpy array of latitude,longitude, and radius
    '''
    radius = np.sqrt(x**2 + y**2 + z**2)
    theta = np.rad2deg(np.arctan2(y,x))
    phi = np.rad2deg(np.arccos(z/radius))
    # lon = (theta + 180) % 360 - 180
    # lon = (theta + 360) % 360
    lon = theta
    lat = 90 - phi
    
    return np.array([lat,lon,radius]).T


def find_match(region_index, region_list):
    ''' 
    Find neighboring regions

    @param region_index: Numeric index of region to find matches for (number between 0 and len(vertices))
    @param region_list: list of lists of vertices that define regions

    @return Numeric indices of regions that border the region specified by region_index
    '''
    regions = region_list[region_index]
    matches = []
    num_matched_list=[]
    
    for i in range(len(region_list)):
        test_regions = region_list[i]

        num_matches = 0
        found = False
        for region in regions:
            if region in test_regions:
                num_matches += 1
                found = True
                
        if found is True:
            matches.append(i)
            
        num_matched_list.append(num_matches)

    return matches


def getVoronoiCollection(data, lat_name, lon_name, bmap = None, v_name = None, full_sphere = False, 
                         max_v=.3, min_v=-0.3, cmap = matplotlib.cm.get_cmap('jet'), test_point = None,
                         proj1=None, proj2=None, **kwargs):
    '''
    Perform a Spherical Voronoi Tessellation on the input data.

    In the case where the data is restricted to one part of the globe, a polygon will not be returned
    for all objects, as matplotlib polygons won't be able to stretch over half the globe. 

    @param data: Input pandas data frame
    @param lat_name: Name of latitude column
    @param lon_name: Name of longitude column
    @param bmap: Basemap instance used to convert from lat, lon coordinates to projection coordinates
    @param v_name: Name of value column. Use this to color each cell according to a value.
    @param full_sphere: Set to true if the data spans the entire globe. 
                        If false, a fictional point is created during tessellation and 
                        removed later to work around issues when polygons are suppose to 
                        span the over half the globe.
    @param max_v: Specify a maximum value to use when assigning values to the tessellation
    @param min_v: Specify a minimum value to use when assigning values to the tessellation
    @param cmap: Matplotlib color map to use
    @param test_point: Tuple containing the latitude and longitude of the ficitonal point to used to remove polygons that
                       wrap around the earth. If none, a point is automatically chosen
    @param proj1: PyProj projection of input coordinates
    @param proj2: PyProj projection of sphere

    @return Matplotlib patch collection of tessellation, scipy.spatial.SphericalVoronoi object, integer index of objects in patch collection.
    '''

    data = data.copy()

    if full_sphere == False:
        if test_point == None:
            test_lat = -1*np.mean(data[lat_name])
            test_lon = np.mean(data[lon_name]) + 180

        else:
            test_lat = test_point[0]
            test_lon = test_point[1]
    
        full_data = data
        full_data = pd.concat([full_data, pd.DataFrame({lat_name: test_lat, lon_name: test_lon}, 
                                                        index=[full_data.index[0]])])
        
        full_data.set_index(np.arange(len(full_data)), inplace=True)

        
    else:
        full_data = data
        
    # print(full_data.tail())

    if proj1 != None and proj2 != None:
        results = pyproj.transform(proj1, proj2, full_data[lon_name].as_matrix(), full_data[lat_name].as_matrix())
        full_data[lon_name] = results[0]
        full_data[lat_name] = results[1]

    xyz = pd.DataFrame(sphericalToXYZ(full_data[lat_name], full_data[lon_name]),columns=['x','y','z'],index=full_data.index)
    
    if v_name != None:
        full_data = pd.concat([full_data.loc[:,[lat_name,lon_name, v_name]],xyz],axis=1)
    else:
        full_data = pd.concat([full_data.loc[:,[lat_name,lon_name, v_name]],xyz],axis=1)
        
    unique_index = np.unique(full_data.loc[:,lat_name] + 1j*full_data.loc[:,lon_name],return_index=True)[1]
    
    full_data = full_data.iloc[np.sort(unique_index)]
    
    voronoi = SphericalVoronoi(full_data.loc[:,['x','y','z']].as_matrix(), **kwargs)
    
    voronoi.sort_vertices_of_regions()
    
    latlon_verts = xyzToSpherical(voronoi.vertices[:,0],voronoi.vertices[:,1], voronoi.vertices[:,2])

    if proj1 != None and proj2 != None:
        results = pyproj.transform(proj2, proj1, latlon_verts[:,1], latlon_verts[:,0])
        latlon_verts[:, 1] = results[0]
        latlon_verts[:, 0] = results[1]
    
    matches = list(map(lambda x: find_match(x, voronoi.regions), range(len(voronoi.regions))))
    
    patch_list = []
    patch_index = []

    for i, (region,match,(station,row)) in enumerate(zip(voronoi.regions,matches,
                                                         full_data.iterrows())):

        if full_sphere or (len(matches)-1) not in match:
            # Keep an index of regions in patchcollection
            patch_index.append(i)

            if bmap != None:
                xy = np.array(bmap(latlon_verts[region,1],latlon_verts[region,0])).T
            else:
                xy = np.array([latlon_verts[region,1],latlon_verts[region,0]]).T
                
            
            if v_name != None:
                value = row[v_name]
                scaled_value = (value - min_v) / (max_v - min_v)
                if scaled_value > 1:
                    scaled_value = 1.0
                elif scaled_value < 0:
                    scaled_value = 0.0

                poly = Polygon(xy, fill=True,facecolor = cmap(scaled_value),edgecolor=cmap(scaled_value))
                
            else:
                poly = Polygon(xy, fill=False)
                
            patch_list.append(poly)

    return matplotlib.collections.PatchCollection(patch_list,match_original=True), voronoi, patch_index



# The MIT License (MIT)
# Copyright (c) 2017 Massachusetts Institute of Technology
#
# Authors: Victor Pankratius, Justin Li, Cody Rude
# This software is part of the NSF DIBBS Project "An Infrastructure for
# Computer Aided Discovery in Geoscience" (PI: V. Pankratius) and 
# NASA AIST Project "Computer-Aided Discovery of Earth Surface 
# Deformation Phenomena" (PI: V. Pankratius)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import time
from collections import OrderedDict

import ipywidgets as widgets
from IPython.display import display

import skdiscovery.utilities.cloud.amazon_control as ac
from skdiscovery.data_structure.framework import config

widget_dict = OrderedDict()
disable_list = ['initialize_button', 'cache_button', 'restore_button']
initialized_disabled_list = ['new_num_instances_widget', 'execute_instances_button']
key_value_list = ['aws_id_widget', 'aws_secret_widget', 'aws_region_widget','aws_security_widget',
                  'aws_keyname_widget','aws_pem_widget','aws_image_id', 'instance_type_widget']

initialized=False


# Creates a GUI for setting up the Amazon instances when offloading pipelines.
# The GUI can just be called directly, taking no inputs or outputs.
def init():
    '''Initialize GUI for controlling Amazon instances'''
    
    global widget_dict


    # Setup Options
    widget_dict['aws_id_widget'] = widgets.Text(
        value=None,
        placeholder='AWS Access ID',
        description='AWS ID:',
        disabled=False,
        width='100%'
    )

    widget_dict['aws_secret_widget'] = widgets.Text(
        value = '',
        placeholder='AWS Secret Access Key',
        description='Secret:',
        width='100%'
    )
    widget_dict['aws_region_widget'] = widgets.Text(
        value = '',
        placeholder='us-west-2',
        description='Region:',
        width='100%'
    )
    widget_dict['aws_security_widget'] = widgets.Text(
        value = '',
        placeholder='Security Group Name',
        description='Security: ',
        width='100%'
    )

    widget_dict['aws_keyname_widget'] = widgets.Text(
        value = '',
        placeholder='Name of Key Pair',
        description='Key Name:',
        width='100%'
    )

    widget_dict['aws_pem_widget'] = widgets.Text(
        value = '',
        placeholder='/home/username/security_key.pem',
        description='PEM File:',
        width='100%'
    )
    
    widget_dict['aws_image_id'] = widgets.Text(
        value = '',
        placeholder='ami-xxxxxxxx',
        description='Image ID:',
        width='100%'
    )    
    widget_dict['instance_type_widget'] = widgets.Text(
        placeholder = 't2.medium',
        description='Type:',
        disabled=False,
        width='100%'
    )

    widget_dict['initialize_button'] = widgets.Button(
        description='Initialize',
        disabled = False,
        button_style='success',
        )

    widget_dict['cache_button'] = widgets.Button(
        description='Save credentials',
        disabled=False,
        button_style='success',        
    )

    widget_dict['restore_button'] = widgets.Button(
        description='Restore credentials',
        disabled=False,
        button_style='success',        
    )

    def cache_cred(b):
        
        changeButtonState(enabled=False)        

        def cache_value(in_widget, key):
            #if in_widget.value != '':
            config.writeConfigValue('AWS', key, in_widget.value)

        cache_value(widget_dict['aws_id_widget'], 'ID')
        cache_value(widget_dict['aws_secret_widget'], 'Secret')
        cache_value(widget_dict['aws_region_widget'], 'Region')
        cache_value(widget_dict['aws_security_widget'], 'Security')
        cache_value(widget_dict['aws_keyname_widget'], 'Key_Name')
        cache_value(widget_dict['aws_pem_widget'], 'PEM_File')
        cache_value(widget_dict['aws_image_id'], 'image_id')
        cache_value(widget_dict['instance_type_widget'],'instance_type')

        changeButtonState(enabled=True)        

    widget_dict['cache_button'].on_click(cache_cred)

    def restore_cred(b):
        changeButtonState(enabled=False)
        conf = config.getConfig()
        widget_dict['aws_id_widget'].value = conf.get('AWS','ID',fallback='')
        widget_dict['aws_secret_widget'].value = conf.get('AWS','Secret',fallback='')
        widget_dict['aws_region_widget'].value = conf.get('AWS','Region',fallback='')
        widget_dict['aws_security_widget'].value = conf.get('AWS','Security',fallback='')
        widget_dict['aws_keyname_widget'].value = conf.get('AWS','Key_Name',fallback='')
        widget_dict['aws_pem_widget'].value = conf.get('AWS', 'PEM_File', fallback='')
        widget_dict['aws_image_id'].value = conf.get('AWS', 'image_id', fallback='')
        widget_dict['instance_type_widget'].value = conf.get('AWS','instance_type',fallback='')
        changeButtonState(enabled=True)        

    widget_dict['restore_button'].on_click(restore_cred)
    # shared_cluster_button = widgets.Button(icon='ban')


    def initialize_cluster(b):
        '''
        Initialize connection to Amazon Web Services and any offloading instances

        @param b: Button that was clicked button
        '''

        global initialized

        changeButtonState(enabled=False)
        widget_dict['aws_status_widget'].value = 'Initializing'

        if checkValidValues():
            ac.init(widget_dict['aws_id_widget'].value,
                    widget_dict['aws_secret_widget'].value,
                    widget_dict['aws_region_widget'].value,
                    widget_dict['aws_security_widget'].value,
                    widget_dict['aws_keyname_widget'].value,
                    widget_dict['aws_pem_widget'].value)

            widget_dict['aws_status_widget'].value = str(len(ac.amazon_list)) +' Amazon node(s) running'
            widget_dict['execute_instances_button'].disabled = False
            widget_dict['initialize_button'].disabled = False
            widget_dict['new_num_instances_widget'].disabled = False
            widget_dict['initialize_button'].description = 'Re-Initialize'
            widget_dict['new_num_instances_widget'].value = len(ac.amazon_list)

            initialized=True

        changeButtonState(enabled=True)


    widget_dict['initialize_button'].on_click(initialize_cluster)
    

    # Controlling Instances
    # Label converted to text    
    widget_dict['label_num_instances'] = widgets.Label(value='Set Number of Instances: ')
    widget_dict['new_num_instances_widget'] = widgets.IntSlider(
        disabled = True,
        description = '',
        value = 0,
        min=0,
        max=10,
        step = 1
    )

    # Label converted to text
    # widget_dict['run_label'] = widgets.Label(value='Status:')
    widget_dict['aws_status_widget'] = widgets.Text(
        value="Not Initialized",
        description='Status:',
        disabled=True,
    )

    widget_dict['execute_instances_button'] = widgets.Button(
        description='Execute',
        disabled=True,
        width='25%'
    )


    def execute_instances(b):
        '''
        Changes number of running instances to number in slider widget

        @param b: Button that was clicked
        '''
        changeButtonState(enabled=False)
        if checkValidValues():
            widget_dict['aws_status_widget'].value = 'Initializing'
            ac.setNumInstances(widget_dict['new_num_instances_widget'].value,
                               widget_dict['instance_type_widget'].value, widget_dict['aws_image_id'].value)
            widget_dict['aws_status_widget'].value = str(len(ac.amazon_list)) +' Amazon node(s) running'
        changeButtonState(enabled=True)

    widget_dict['execute_instances_button'].on_click(execute_instances)

    restore_cred(None)
    drawGUI()
    
def drawGUI():
    '''
    Draw the GUI on the screen
    '''
    display(widget_dict['aws_id_widget'])
    display(widget_dict['aws_secret_widget'])
    display(widget_dict['aws_region_widget'])
    display(widget_dict['aws_security_widget'])
    display(widget_dict['aws_keyname_widget'])
    display(widget_dict['aws_pem_widget'])
    display(widget_dict['aws_image_id'])    
    display(widget_dict['instance_type_widget'])
    display(widgets.HBox([widget_dict['initialize_button'], widget_dict['cache_button'],widget_dict['restore_button']]))

    # display(widgets.HBox([widget_dict['label_num_instances'], widget_dict['new_num_instances_widget']]))
    # display(widgets.HBox([widget_dict['run_label'], widget_dict['aws_status_widget']]))

    display(widgets.HBox([widget_dict['label_num_instances'], widget_dict['new_num_instances_widget']]))
    display(widget_dict['aws_status_widget'])
    display(widget_dict['execute_instances_button'])
    # Need to disable after displaying in ipywidgets 7
    widget_dict['execute_instances_button'].disabled=True
    widget_dict['execute_instances_button'].button_style=''
    widget_dict['new_num_instances_widget'].disabled=True


def changeButtonState(enabled=True):
    '''
    Enable or disable the buttons and slider in the GUI

    @param enabled: State to change the buttons to.
    '''
    if enabled==True:
        new_disabled = False
        button_style = 'success'
    else:
        new_disabled = True
        button_style = ''
        
    global widget_dict
    global disable_list
    for label in disable_list:
        widget_dict[label].disabled = new_disabled
        widget_dict[label].button_style = button_style

    if initialized:
        for label in initialized_disabled_list:
            widget_dict[label].disabled = new_disabled
            widget_dict[label].button_style = button_style

        
def checkValidValues():
    '''
    Check if Amazon information is valid

    @return True if all AWS text fields have data in them, false otherwise
    '''
    is_valid = True
    for key in key_value_list:
        if widget_dict[key].value == '':
            is_valid = False
            break

    if is_valid == False:
        prev_message = widget_dict['aws_status_widget'].value
        widget_dict['aws_status_widget'].value = 'Please fill in all information'
        time.sleep(5)

    return is_valid



from .base import PipelineItem
# from .param import *
from .discoverypipeline import DiscoveryPipeline
from .stagecontainers import *
# from .base_region_detect import BaseDetector
# from .base_image_classify import BaseClassifier
